import os
import json
import datetime
from pupa.core import db
from pupa.models import Organization
from pupa.models.utils import DatetimeValidator
from pupa.models.schemas.jurisdiction import schema as jurisdiction_schema


def import_jurisdiction(org_importer, jurisdiction):
    obj = jurisdiction.get_db_object()

    obj['_type'] = 'jurisdiction'
    obj['_id'] = jurisdiction.jurisdiction_id

    if not obj['_id'].startswith("ocd-jurisdiction/"):
        raise ValueError("The Jurisdiction appears to have an ID that does not"
                         " begin with 'ocd-jurisdiction'. I found '%s'" % (
                         jurisdiction.jurisdiction_id))

    obj['latest_update'] = datetime.datetime.utcnow()

    # validate jurisdiction
    validator = DatetimeValidator()
    try:
        validator.validate(obj, jurisdiction_schema)
    except ValueError as ve:
        raise ve

    db.jurisdictions.save(obj)

    # create organization(s)
    org = Organization(name=jurisdiction.name, classification='legislature',
                       jurisdiction_id=jurisdiction.jurisdiction_id)
    if jurisdiction.other_names:
        org.other_names = jurisdiction.other_names
    if jurisdiction.parent_id:
        org.parent_id = jurisdiction.parent_id

    parent_id = org_importer.import_object(org)

    if jurisdiction.chambers:
        for chamber, properties in jurisdiction.chambers.items():
            org = Organization(name=properties['name'], classification='legislature',
                               chamber=chamber, parent_id=parent_id,
                               jurisdiction_id=jurisdiction.jurisdiction_id)
            org_importer.import_object(org)

    # create parties
    for party in jurisdiction.parties:
        org = Organization(**{'classification': 'party',
                              'name': party['name'],
                              'parent_id': None})
        org_importer.import_object(org)



from pupa.models import Vote
from .utils import (people_by_jurisdiction_and_name,
                    orgs_by_jurisdiction_and_name,
                    bills_by_jurisdiction_and_name)
from .base import BaseImporter


class VoteImporter(BaseImporter):
    _type = 'vote'
    _model_class = Vote

    def __init__(self, jurisdiction_id,
                 person_importer, org_importer, bill_importer):

        super(VoteImporter, self).__init__(jurisdiction_id)
        self.person_importer = person_importer
        self.bill_importer = bill_importer
        self.org_importer = org_importer

    def get_db_spec(self, vote):
        spec = {
            "motion": vote.motion,
            "chamber": vote.chamber,
            "date": vote.date,
            "jurisdiction_id": vote.jurisdiction_id,
        }
        return spec

    def prepare_object_from_json(self, obj):
        bill = obj.get('bill', None)
        if bill:
            if bill.get('id'):
                # We've been given a hard ID. Let's use it to match against
                # the real bill. (since the scraper knew what they were doing)
                bill['id'] = self.bill_importer.resolve_json_id(bill['id'])
            else:
                # OK. Right. We weren't given an ID. Let's try to do a
                # match by name.
                bills = bills_by_jurisdiction_and_name(
                    obj['jurisdiction_id'],
                    bill['name'],
                )
                if bills.count() != 1:
                    self.warning("Can't resolve bill `%s'" % (bill['name']))
                else:
                    bill_obj = bills[0]
                    bill['id'] = bill_obj['_id']

        for vote in obj['roll_call']:
            who = vote['person']
            people = people_by_jurisdiction_and_name(
                obj['jurisdiction_id'],
                who['name'],
                chamber=who.get('chamber')
            )

            if people.count() != 1:
                self.warning("can't match `%s'" % (who['name']))
                continue  # can't match

            person_obj = people[0]
            vote['person']['id'] = person_obj['_id']

        org = obj.get('organization')
        org_id = obj.get('organization_id')

        if org and not org_id:  # OK. We have an org that needs matching.
            orgs = orgs_by_jurisdiction_and_name(
                obj['jurisdiction_id'],
                org,
            )  # get all matching orgs.

            if orgs.count() == 1:
                org_obj = orgs[0]  # Let's get the only result.
                obj['organization_id'] = org_obj['_id']
            else:
                self.warning("can't match `%s'" % (org))

        elif org_id:  # We have a sort of org ID
            if org is None:  # If we have the ID but no the name (odd...)
                raise ValueError("Someone set an org_id without an org name.")

            org_json_id = obj['organization_id']  # scrape-time match?
            if org_json_id and not org_json_id.startswith("ocd-organization"):
                obj['organization_id'] = self.org_importer.resolve_json_id(
                    org_json_id)  # resolve it.
        return obj



from pupa.core import db
from pupa.models import Organization
from .base import BaseImporter


class OrganizationImporter(BaseImporter):
    _type = 'organization'
    _model_class = Organization

    def get_db_spec(self, org):
        spec = {'classification': org.classification,
                'name': org.name,
                'parent_id': org.parent_id}

        if org.classification not in ("party", ):  # just party for now
            spec['jurisdiction_id'] = org.jurisdiction_id

        return spec

    def _resolve_org_by_chamber(self, jurisdiction_id, chamber):
        """
        This is used by the Bill importer to match an org purely based on
        ``chamber`` if it exists.
        """

        orgs = db.organizations.find({
            "classification": "legislature",
            "jurisdiction_id": jurisdiction_id,
            "chamber": chamber
        })

        if orgs.count() == 1:
            return orgs[0]  # Neato! We found one!
        elif orgs.count() == 0:
            raise ValueError("Chamber `%s' isn't giving us an org in `%s'" % (
                chamber, jurisdiction_id
            ))
        else:
            raise ValueError("Chamber `%s' isn't a unique org in `%s'" % (
                chamber, jurisdiction_id
            ))

    def resolve_json_id(self, json_id):
        # handle special party:* and jurisdiction:* ids first
        for type_, key in (('party', 'name'),
                           ('jurisdiction', 'jurisdiction_id')):
            if json_id.startswith(type_ + ':'):
                id_piece = json_id.split(':', 1)[1]
                org = db.organizations.find_one(
                    {'classification': type_, key: id_piece})
                if not org:
                    raise ValueError('attempt to create membership to unknown '
                                     + type_ + ': ' + id_piece)
                else:
                    return org['_id']

        # just resolve the normal way
        return super(OrganizationImporter, self).resolve_json_id(json_id)



from .base import BaseModel
from .schemas.membership import schema


class Membership(BaseModel):
    """
    A single popolo encoded Membership.
    """

    _type = 'membership'
    _schema = schema
    _collection = 'memberships'
    __slots__ = ("organization_id", "person_id", "post_id", "role",
                 "start_date", "end_date", "contact_details", "chamber",
                 'jurisdiction_id', 'unmatched_legislator', 'label',
                 'division_id', 'on_behalf_of_id', 'links',
                )

    def __init__(self, person_id, organization_id, **kwargs):
        """
        Constructor for the Person object.

        We require a person ID and organization ID, as required by the
        popolo spec. Additional arguments may be given, which match those
        defined by popolo.
        """
        super(Membership, self).__init__()
        self.person_id = person_id
        self.organization_id = organization_id
        self.contact_details = []
        self.start_date = None
        self.end_date = None
        self.post_id = None
        self.role = None
        self.unmatched_legislator = None
        self.label = None
        self.on_behalf_of_id = None
        self.links = []
        self.division_id = None

        for k, v in kwargs.items():
            setattr(self, k, v)

    def add_contact_detail(self, type, value, note):
        self.contact_details.append({"type": type,
                                     "value": value,
                                     "note": note})

    def add_link(self, url, note=None):
        self.links.append({"note": note, "url": url})

    def __str__(self):
        if self.person_id:
            return self.person_id + ' membership in ' + self.organization_id
        else:
            return (self.unmatched_legislator['name'] + ' membership in ' +
                    self.organization_id)
    __unicode__ = __str__



""" these are helper classes for object creation during the scrape """
from pupa.models.person import Person
from pupa.models.organization import Organization
from pupa.models.membership import Membership


class Legislator(Person):
    _is_legislator = True
    __slots__ = ('post_id', 'party', 'chamber', '_contact_details', '_role')

    def __init__(self, name, post_id, party=None, chamber=None, role='member', **kwargs):
        super(Legislator, self).__init__(name, **kwargs)
        self.post_id = post_id
        self.party = party
        self.chamber = chamber
        self._contact_details = []
        self._role = role

    def add_contact(self, type, value, note):
        self._contact_details.append({'type': type, 'value': value,
                                      'note': note})

    def add_committee_membership(self, com_name, role='member'):
        org = Organization(com_name, classification='committee')
        self.add_membership(org, role=role)
        org.sources = self.sources
        self._related.append(org)


class Committee(Organization):

    def __init__(self, *args, **kwargs):
        super(Committee, self).__init__(*args, **kwargs)

    def add_member(self, name, role='member', **kwargs):
        membership = Membership(None, self._id, role=role,
                                unmatched_legislator={'name': name},
                                **kwargs)
        self._related.append(membership)



import uuid
from .utils import DatetimeValidator
from pupa.core import db


class BaseModel(object):
    """
    This is the base class for all the Open Civic objects. This contains
    common methods and abstractions for OCD objects.
    """

    # needs slots defined so children __slots__ are enforced
    __slots__ = ('_id', '_related', 'sources', 'created_at', 'updated_at',
                 'extras', '_meta')

    # to be overridden by children. Something like "person" or "organization".
    # Used in :func:`validate`.
    _type = None
    _schema = None

    def __init__(self):
        self.extras = {}
        self._id = str(uuid.uuid1())
        self.sources = []
        self._meta = {}

    # scraping helpers

    def add_extra(self, key, value):
        self.extras[key] = value

    def add_meta(self, key, value):
        self._meta[key] = value

    def add_meta_software(self, name):
        self.add_meta('software', name)

    def add_source(self, url, note=None, **kwargs):
        """ Add a source URL from which data was collected """
        new = kwargs.copy()
        new.update({'url': url, 'note': note})
        self.sources.append(new)

    # validation

    def validate(self, schema=None):
        """
        Validate that we have a valid object.

        On error, this will either raise a `ValueError` or a
        `validictory.ValidationError` (a subclass of `ValueError`).

        This also expects that the schemas assume that omitting required
        in the schema asserts the field is optional, not required. This is
        due to upstream schemas being in JSON Schema v3, and not validictory's
        modified syntax.
        """
        if schema is None:
            schema = self._schema

        validator = DatetimeValidator(required_by_default=False)
        validator.validate(self.as_dict(), schema)

    # from/to dict

    @classmethod
    def from_dict(cls, db_obj):
        db_obj = db_obj.copy()
        _type = db_obj.pop('_type')

        # hack to get location set correctly since constructor mangles it
        if _type == 'event':
            location = db_obj.pop('location')
            db_obj['location'] = None

        if _type == 'vote':
            # Hack to deal with Vote mangling. We abstract away the 
            # counts so that scrapers don't have to deal with it. This
            # isn't great, but kinda needed.
            counts = db_obj['vote_counts']
            for count in counts:
                db_obj[{
                    "yes": "yes_count",
                    "no": "no_count",
                    "other": "other_count"
                }[count['vote_type']]] = count['count']

        try:
            newobj = cls(**db_obj)
        except TypeError:
            # OK. Let's try to save some good debug information.
            print("Error: Signature doesn't match object on disk.")
            print("We're trying to create a: %s" % (cls.__name__))
            print("Dump of keys from object:")
            for key in db_obj.keys():
                print("  %s" % (key))
            print("This can mean that you have a stale scraped_data folder")
            print("Please remove it and re-try the scrape.")
            print("")
            print("Object in question:")
            print("  %s" % (db_obj['_id']))
            print("")
            raise

        # set correct location object after creation
        if _type == 'event':
            newobj.location = location

        return newobj

    def as_dict(self):
        d = {}
        all_slots = set(self.__slots__)
        for cls in self.__class__.__mro__:
            all_slots |= set(cls.__slots__)
            if cls == BaseModel:
                break
        for attr in all_slots:
            if attr != '_related' and hasattr(self, attr):
                d[attr] = getattr(self, attr)
        d['_type'] = self._type
        return d

    # database stuff
    def save(self):
        db[self._collection].save(self.as_dict())


    # operators

    def __eq__(self, other):
        """ equality requires all fields to be equal except for the _id """
        sd = self.as_dict()
        od = other.as_dict()
        sd.pop('_id')
        od.pop('_id')
        return sd == od




from __future__ import print_function
from contextlib import contextmanager
import json
import os
from pupa.core import db
from .base import BaseCommand
from pupa.utils import JSONEncoderPlus


@contextmanager
def cd(path):
    pop = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(pop)


def normalize_person(entry):
    data = list(db.memberships.find({
        "person_id": entry['_id']
    }, timeout=False))
    for datum in data:
        datum.pop('_id')

    entry['memberships'] = data

    return entry


def dump(collection, spec):
    for entry in collection.find(spec, timeout=False):
        do_write(entry)


def do_write(entry, where=None):
    path = entry['_id']

    if where is None:
        where = entry['jurisdiction_id']

    path = "%s/%s" % (where, path)
    basename = os.path.dirname(path)
    if not os.path.exists(basename):
        os.makedirs(basename)

    with open(path, 'w') as fd:
        #print path
        json.dump(entry, fd, cls=JSONEncoderPlus)


def dump_people(where):
    iterated = False
    for orga in db.organizations.find({"jurisdiction_id": where,
                                       "classification": "legislature"}):
        iterated = True
        for membership in db.memberships.find({"organization_id": orga['_id']},
                                              timeout=False):
            person = db.people.find_one({"_id": membership['person_id']})
            assert person is not None
            person = normalize_person(person)
            do_write(person, where=where)

    if iterated is False:
        raise Exception("Org came back none for %s" % (where))


def dump_jurisdiction_data(where):
    meta = db.jurisdictions.find_one({"_id": where})
    if meta is None:
        return

    for x in [
        'latest_json_url', 'latest_csv_url',
        'latest_json_date', 'latest_csv_date'
    ]:
        if x in meta:
            meta.pop(x, None)

    for key in meta.keys():
        if key.startswith("_") and key != "_id":
            meta.pop(key)

    path = "%s/jurisdiction.json" % (meta['_id'])
    with open(path, 'w') as fd:
        #print path
        json.dump(meta, fd, cls=JSONEncoderPlus)


def dump_juris(jurisdiction):
    spec = {"jurisdiction_id": jurisdiction}

    for collection in [db.bills, db.votes, db.events, db.organizations]:
        dump(collection, spec)

    dump_people(jurisdiction)
    dump_jurisdiction_data(jurisdiction)


class Command(BaseCommand):
    name = 'dump'
    help = '''create a json dump'''

    def add_args(self):
        self.add_argument('jurisdiction', type=str, help='jurisdiction_id to dump',
                          default=None, nargs='?')
        self.add_argument('--output', type=str, help='output directory', default="dump")

    def handle(self, args):
        path = args.output
        if not os.path.exists(path):
            os.makedirs(path)

        with cd(path):
            if args.jurisdiction:
                dump_juris(args.jurisdiction)
            else:
                for orga in db.organizations.find({
                    "classification": "legislature",
                }, timeout=False):
                    if 'jurisdiction_id' not in orga:
                        print("WARNING: NO JURISDICTION_ID ON %s" % (orga['_id']))
                        continue

                    dump_juris(orga['jurisdiction_id'])



from .base import BaseModel
from .utils import add_associated_link
from .schemas.event import schema


class EventAgendaItem(dict):
    event = None
    _add_associated_link = add_associated_link

    def __init__(self, description, event):
        super(EventAgendaItem, self).__init__({
            "description": description,
            "related_entities": [],
            "subjects": [],
            "media": [],
            "notes": [],
            "order": None,
        })
        self.event = event

    def add_subject(self, what):
        self['subjects'].append(what)

    def add_committee(self, committee, id=None, note='participant'):
        self.add_entity(name=committee, type='committee', id=id, note=note)

    def add_bill(self, bill, id=None, note='consideration'):
        self.add_entity(name=bill, type='bill', id=id, note=note)

    def add_person(self, person, id=None, note='participant'):
        self.add_entity(name=person, type='person', id=id, note=note)

    def add_media_link(
        self, name, url, type='media',
        mimetype=None,
        offset=None,
        on_duplicate='error'
    ):
        return self._add_associated_link(
            collection='media',
            name=name,
            url=url,
            type=type,
            offset=offset,
            mimetype=mimetype,
            on_duplicate=on_duplicate)

    def add_entity(self, name, type, id, note):
        self['related_entities'].append({
            "name": name,
            "type": type,
            "id": id,
            "note": note,
        })


class Event(BaseModel):
    """
    Details for an event in .format
    """
    _type = 'event'
    _schema = schema
    _collection = 'events'
    __slots__ = ("when", "all_day", "name", "description", "documents",
                 "end", "links", "location", "participants",
                 "agenda", "sources", "status", "type", 'session',
                 "media", '_openstates_id', 'jurisdiction_id',
                 'identifiers',)

    def __init__(self, name, when, location, session=None, **kwargs):
        super(Event, self).__init__()
        self.when = when
        self.name = name
        self.all_day = False
        self.documents = []
        self.description = None
        self.end = None
        self.links = []
        self.location = {"name": location,
                         "note": None,
                         "coordinates": None}
        self.participants = []
        self.media = []
        self.agenda = []
        self.sources = []
        self.status = "confirmed"
        self.type = "event"
        self._related = []
        self.session = session

        for k, v in kwargs.items():
            setattr(self, k, v)

    def __str__(self):
        return u'{0} {1}'.format(self.when, self.name.strip())
    __unicode__ = __str__

    def set_coordinates(self, lat, lon):
        self.location['coordinates'] = {
            "latitude": lat,
            "longitude": lon
        }

    def add_location_url(self, url):
        self.location['url'] = url

    def add_source(self, url, note=None):
        info = {"url": url, "note": note}
        self.sources.append(info)

    def add_link(self, url, note=None):
        info = {"url": url, "note": note}
        self.links.append(info)

    def add_document(self, name, url, mimetype, **kwargs):
        data = kwargs.copy()
        data.update({
            "name": name,
            "url": url,
            "mimetype": mimetype,
        })
        self.documents.append(data)

    def add_person(self, name, note='participant', chamber=None, id=None):
        return self.add_participant(
            name=name,
            type='person',
            chamber=chamber,
            note=note)

    def add_participant(self, name, type, note='participant', chamber=None,
                        id=None):
        self.participants.append({
            "chamber": chamber,
            "type": type,
            "note": note,
            "name": name,
            "id": id,
        })

    def add_agenda_item(self, description):
        obj = EventAgendaItem(description, self)
        self.agenda.append(obj)
        return obj

    _add_associated_link = add_associated_link

    def add_media_link(
        self, name, url, type='media',
        mimetype=None,
        offset=None,
        on_duplicate='error'
    ):
        return self._add_associated_link(
            collection='media',
            name=name,
            url=url,
            type=type,
            offset=offset,
            mimetype=mimetype,
            on_duplicate=on_duplicate)



from pupa.models.organization import Organization


class Jurisdiction(object):
    """ Base class for a jurisdiction """

    # schema objects
    name = None
    url = None
    chambers = {}
    terms = None
    session_details = None
    feature_flags = []
    building_maps = []
    provides = []
    parties = []
    other_names = []
    parent_id = None
    _ignored_scraped_sessions = []

    _party_cache = {}

    def get_db_object(self):
        return {'name': self.name,
                'url': self.url,
                'chambers': self.chambers,
                'terms': self.terms,
                'session_details': self.session_details,
                'feature_flags': self.feature_flags,
                'building_maps': self.building_maps}

    def term_for_session(self, session):
        for term in self.terms:
            if session in term['sessions']:
                return term['name']
        raise ValueError('no such session: ' + session)

    def get_term_details(self, termname):
        for term in self.terms:
            if term['name'] == termname:
                return term
        raise ValueError('no such term: ' + termname)

    def get_party(self, party_name):
        if not self._party_cache:
            for party in self.parties:
                self._party_cache[party['name']] = Organization(
                    party['name'], _id=party['id'])
        try:
            return self._party_cache[party_name]
        except KeyError:
            raise ValueError('no such party: ' + party_name)

    def get_session_list(self):
        raise NotImplementedError('get_session_list is not implemented')

    def extract_text(self):
        raise NotImplementedError('extract_text is not implemented')

    def get_scraper(self, term, session, obj_type):
        raise NotImplementedError('get_scraper is not implemented')



from .base import BaseModel
from .membership import Membership
from .schemas.person import schema


class Person(BaseModel):
    """
    Details for a Person in Popolo format.
    """

    _type = 'person'
    _schema = schema
    _collection = 'people'

    __slots__ = ('name', 'gender', 'birth_date',
                 'death_date', 'image', 'summary', 'biography', 'links',
                 'other_names', 'contact_details', '_openstates_id',
                 'chamber', 'district', 'identifiers',
                 'post_id', 'national_identity',
                )
    _other_name_slots = ('name', 'start_date', 'end_date', 'note')

    def __init__(self, name, **kwargs):
        super(Person, self).__init__()
        self.name = name
        self.biography = None
        self.summary = None
        self.birth_date = None
        self.death_date = None
        self.image = None
        self.gender = None
        self.national_identity = None
        self.links = []
        self.identifiers = []
        self.other_names = []
        self._related = []
        self.contact_details = []

        for k, v in kwargs.items():
            setattr(self, k, v)

    def add_name(self, name, **kwargs):
        other_name = {'name': name}
        for k, v in kwargs.items():
            if k not in self._other_name_slots:
                raise AttributeError('{0} not a valid kwarg for add_name'
                                     .format(k))
            other_name[k] = v
        self.other_names.append(other_name)

    def add_link(self, url, note=None):
        self.links.append({"note": note, "url": url})

    def add_identifier(self, identifier, scheme=None):
        data = {"identifier": identifier}
        if scheme:
            data['scheme'] = scheme
        self.identifiers.append(data)

    def add_contact_detail(self, type, value, note):
        self.contact_details.append({"type": type,
                                     "value": value,
                                     "note": note})

    def add_membership(self, organization, role='member', **kwargs):
        """
            add a membership in an organization and return the membership
            object in case there are more details to add
        """
        membership = Membership(self._id, organization._id, role=role,
                                **kwargs)
        self._related.append(membership)
        return membership

    def __str__(self):
        return self.name
    __unicode__ = __str__

    def __repr__(self):
        as_dict = self.as_dict()
        list(map(as_dict.pop, ('_type', '_id', 'name')))
        args = (self.__class__.__name__, self.name, as_dict)
        return '%s(name=%r, **%r)' % args



from __future__ import print_function
import datetime as dt
import uuid
import sys
from pymongo import Connection
from .base import BaseCommand
from pupa.core import db
from pupa.models import Organization, Membership, Person, Event, Vote, Bill

import csv

indices = [
    ("bills", "_openstates_id"),
    ("people", "_openstates_id"),
    ("organizations", "_openstates_id"),
    ("memberships", "_openstates_id"),
    ("events", "_openstates_id"),
    ("votes", "_openstates_id"),
]


type_tables = {
    Organization: "organizations",
    Membership: "memberships",
    Person: "people",
    Bill: "bills",
    Event: "events",
    Vote: "votes",
}

_hot_cache = {}
_cache_touched = {}
name_cache = {}


def obj_to_jid(obj):
    if obj._openstates_id:
        jid = openstates_to_jid(obj._openstates_id)

    if jid is None:
        raise Exception("Can't auto-detect the Jurisdiction ID")

    return jid


def openstates_to_jid(osid):
    abbr = osid[:2]
    abbr = abbr.lower()
    jid = "ocd-jurisdiction/country:us/state:%s/legislature" % (abbr)
    return jid


def load_hot_cache(state):
    spec = {}
    if state:
        spec['state'] = state
    for entry in db.openstates_cache.find(spec, timeout=False):
        _hot_cache[entry['_id']] = entry['ocd-id']


def write_hot_cache(state):
    global _cache_touched
    for entry in _hot_cache:
        if entry is None:
            continue

        if _cache_touched.get(entry, False):
            db.openstates_cache.update({"_id": entry}, {"_id": entry, "state": entry[:2].lower(),
                                                        "ocd-id": _hot_cache[entry]}, upsert=True,
                                       safe=True)
    _cache_touched = {}


def ocd_namer(obj):
    # ocd-person/UUID
    # ocd-organization/UUID
    if hasattr(obj, '_openstates_id'):
        ret = _hot_cache.get(obj._openstates_id)
        if ret is not None:
            return ret

        # OK. Let's try a last-ditch.
        what = type_tables[type(obj)]
        ## This is one of the *major* time-sinks. This codepath is taken when
        # the last run only went halfway, without writing the cache back out.
        dbobj = getattr(db, what).find_one({
            "_openstates_id": obj._openstates_id
        })
        if dbobj:
            _hot_cache[dbobj['_openstates_id']] = dbobj['_id']
            return dbobj['_id']

    if obj._type == 'membership':
        return None

    return "ocd-{type_}/{uuid}".format(type_=obj._type, uuid=uuid.uuid1())


def is_ocd_id(string):
    return string.startswith("ocd-")


def lookup_entry_id(collection, openstates_id):
    if openstates_id is None:
        return None

    hcid = _hot_cache.get(openstates_id, None)
    if hcid:
        return hcid

    org = getattr(db, collection).find_one({
        "_openstates_id": openstates_id
    })

    if org is None:
        return None

    id_ = str(org['_id'])
    _hot_cache[openstates_id] = id_
    _cache_touched[openstates_id] = True
    return id_


def get_current_term(jid):
    meta = get_metadata(jid)
    term = meta['terms'][-1]
    return term


def get_metadata(what):
    hcid = _hot_cache.get(what, None)
    if hcid:
        return hcid

    meta = db.jurisdictions.find_one({"_id": what})
    _hot_cache[what] = meta
    return meta


class Command(BaseCommand):
    name = 'import-billy'
    help = '''import data from a billy instance'''

    def add_args(self):
        self.add_argument('mappings', type=str, help='OpenStates mapping files to use')
        self.add_argument('ocd_mapping', type=str, help='OpenCivic mapping file to use (country-us.csv)')
        self.add_argument('state', type=str, help='State to rebuild', default=None, nargs='?')
        self.add_argument('--billy-server', type=str, help='Billy Mongo Server',
                          default="localhost")
        self.add_argument('--billy-database', type=str, help='Billy Mongo Database',
                          default="fiftystates")
        self.add_argument('--billy-port', type=int, help='Billy Mongo Server Port', default=27017)
        self.add_argument('--quiet', action='store_false', help='Dont spam my screen',
                          default=True)
        self.add_argument('--dont-validate', action='store_false',
                          help='Dont validate incoming objects', default=True)

    def handle(self, args):
        state = args.state
        self.quiet = args.quiet
        self.validate = args.dont_validate
        self.billy_db = Connection(args.billy_server, args.billy_port)[args.billy_database]
        self.mapping_dir = args.mappings
        self.ocd_mapping = args.ocd_mapping

        self.valid_geo_ids = {}
        self.osid_to_geo_id = {}
        for row in csv.DictReader(open(self.ocd_mapping, 'r').readlines()):
            self.valid_geo_ids[row['id']] = row

            if row['openstates_district'] != "":
                self.osid_to_geo_id[row['openstates_district']] = row

        for database, index in indices:
            db[database].ensure_index(index)

        SEQUENCE = [
            load_hot_cache,
            self.migrate_legislatures,
            self.migrate_people,  # depends on legislatures
            #
            # XXX: migrate_people needs to be called for migrate_committees, for two
            #      reasons:
            #
            #   - committess look up person _id entries
            #   - migrating people drops memberships, which means it'd avoid
            #     dupes.
            #
            self.migrate_committees,  # depends on people
            self.migrate_bills,
            self.migrate_events,
            self.migrate_votes,
            write_hot_cache,
        ]

        def handle_state(state):
            print("now processing", state)
            for seq in SEQUENCE:
                seq(state)

        if state:
            handle_state(state)
        else:
            for state in (x['abbreviation'] for x in self.billy_db.metadata.find(timeout=False)):
                handle_state(state)

        print("\nMigration complete.")

    def create_or_get_party(self, what):
        hcid = _hot_cache.get(what, None)
        if hcid:
            return hcid

        org = db.organizations.find_one({
            "name": what
        })
        if org:
            _hot_cache[what] = org['_id']
            _cache_touched[what] = True
            return org['_id']

        org = Organization(what, classification="party")

        self.save_object(org)

        _hot_cache[what] = org._id
        _cache_touched[what] = True

        return org._id

    def migrate_legislatures(self, state):
        spec = {}
        if state:
            spec['_id'] = state

        for metad in self.billy_db.metadata.find(spec, timeout=False):
            abbr = metad['abbreviation']
            geoid = "ocd-division/country:us/state:%s" % (abbr)
            for chamber in metad['chambers']:
                cn = metad['chambers'][chamber]['name']
                cow = Organization("%s, %s" % (metad['legislature_name'], cn),
                                   classification="legislature",
                                   chamber=chamber,
                                   division_id=geoid,
                                   abbreviation=abbr)
                cow._openstates_id = "%s-%s" % (abbr, chamber)
                cow.add_source(metad['legislature_url'])

                for post in self.billy_db.districts.find({"abbr": abbr}):
                    if post['chamber'] != chamber:
                        continue

                    cow.add_post(label="Member", role="member", num_seats=post['num_seats'],
                                 id=post['name'])

                self.save_object(cow)

            meta = self.billy_db.metadata.find_one({"_id": cow.abbreviation})
            if meta is None:
                raise Exception
            meta.pop("_id")
            meta['_id'] = cow.jurisdiction_id

            for badtag in ["latest_json_url", "latest_json_date",
                           "latest_csv_url", "latest_csv_date"]:
                meta.pop(badtag, None)

            meta['division_id'] = "ocd-division/country:us/state:%s" % (
                cow.abbreviation
            )

            db.jurisdictions.save(meta)

    def migrate_committees(self, state):

        def attach_members(committee, org):
            term = get_current_term(obj_to_jid(org))
            for member in committee['members']:
                osid = member.get('leg_id', None)
                person_id = lookup_entry_id('people', osid)
                if person_id:
                    m = Membership(person_id, org._id,
                                   role=member['role'],
                                   chamber=org.chamber,
                                   # term=term['name'],
                                   start_date=str(term['start_year']))
                    m.add_extra('term', term['name'])
                    # We can assume there's no end_year because it's a current
                    # member of the committee. If they left the committee, we don't
                    # know about it yet :)
                    self.save_object(m)

                    if m.role != 'member':
                        # In addition to being the (chair|vice-chair),
                        # they should also be noted as a member.
                        m = Membership(person_id, org._id,
                                       role='member',
                                       chamber=org.chamber,
                                       start_date=str(term['start_year']))
                        m.add_extra('term', term['name'])
                        self.save_object(m)

        spec = {"subcommittee": None}

        if state:
            spec['state'] = state

        for committee in self.billy_db.committees.find(spec, timeout=False):
            # OK, we need to do the root committees first, so that we have IDs that
            # we can latch onto down below.
            org = Organization(committee['committee'],
                               classification="committee")
            org.chamber = committee['chamber']
            org.parent_id = lookup_entry_id('organizations', committee['state'])
            org.identifiers = [{'scheme': 'openstates',
                                'identifier': committee['_id']}]
            org._openstates_id = committee['_id']
            org.sources = committee['sources']
            org.created_at = committee['created_at']
            org.updated_at = committee['updated_at']
            # Look into posts; but we can't be sure.
            self.save_object(org)
            attach_members(committee, org)

        spec.update({"subcommittee": {"$ne": None}})

        for committee in self.billy_db.committees.find(spec, timeout=False):
            org = Organization(committee['subcommittee'],
                               classification="committee")

            org.parent_id = lookup_entry_id(
                'organizations',
                committee['parent_id']
            ) or lookup_entry_id(
                'organizations',
                committee['state']
            )

            org.identifiers = [{'scheme': 'openstates',
                               'identifier': committee['_id']}]
            org._openstates_id = committee['_id']
            org.sources = committee['sources']
            org.chamber = committee['chamber']
            # Look into posts; but we can't be sure.
            self.save_object(org)
            attach_members(committee, org)

    def migrate_people(self, state):
        spec = {}
        if state:
            spec["state"] = state
        for entry in self.billy_db.legislators.find(spec, timeout=False):
            jurisdiction_id = openstates_to_jid(entry['_id'])

            who = Person(entry['full_name'])
            who.identifiers = [{'scheme': 'openstates',
                               'identifier': entry['_id']}]
            who._openstates_id = entry['_id']
            who.created_at = entry['created_at']
            who.updated_at = entry['updated_at']
            if who.name != entry['_scraped_name']:
                who.add_name(entry['_scraped_name'])

            for k, v in {
                "photo_url": "image",
                "chamber": "chamber",
                "district": "district",
            }.items():
                if entry.get(k, None):
                    setattr(who, v, entry[k])

            who.sources = entry['sources']

            home = entry.get('url', None)
            if home:
                who.add_link(home, "Homepage")

            vsid = entry.get('votesmart_id')
            if vsid:
                who.add_identifier(vsid, 'votesmart-id')

            tdid = entry.get('transparencydata_id')
            if tdid:
                who.add_identifier(tdid, 'transparencydata-id')

            blacklist = ["photo_url", "chamber", "district", "url",
                         "roles", "offices", "party", "state", "sources",
                         "active", "old_roles", "_locked_fields", "created_at",
                         "updated_at", "transparencydata_id", "votesmart_id",
                         "leg_id", "email", "phone", "fax", "_scraped_name",
                         "_id", "_all_ids", "full_name", "country", "level",
                         "office_address", "suffixes", "_type"]
            # The above is a list of keys we move over by hand later.

            for key, value in entry.items():
                if key in blacklist or not value:  # or key.startswith("_"):
                    continue
                who.extras[key] = value

            who.add_meta('locked_fields', entry.get('_locked_fields', []))

            chamber = entry.get('chamber')
            if chamber == "joint":
                continue

            legislature = None
            if chamber:
                legislature = _hot_cache.get('{state}-{chamber}'.format(
                    state=entry['state'],
                    chamber=chamber,
                ))

                if legislature is None:
                    print(chamber, entry['state'], entry['_id'])
                    raise Exception("Someone's in the void.")

            self.save_object(who)  # gives who an id, btw.

            party = entry.get('party', None)
            db.memberships.remove({"person_id": who._id}, safe=True)

            if party:
                m = Membership(who._id, self.create_or_get_party(entry['party']))
                self.save_object(m)

            if legislature:
                term = get_current_term(jurisdiction_id)
                m = Membership(who._id, legislature,
                               chamber=chamber,
                               start_date=str(term['start_year']))
                m.add_extra('term', term['name'])

                chamber, district = (entry.get(x, None)
                                     for x in ['chamber', 'district'])

                if chamber:
                    m.chamber = chamber

                if district:
                    m.post_id = district

                for key in ['email', 'fax', 'phone']:
                    if key in entry and entry[key]:
                        m.add_contact_detail(type=key,
                                             value=entry[key],
                                             note=key)

                if entry.get("office_address"):
                    m.add_contact_detail(type='office',
                                         value=entry['office_address'],
                                         note='Office Address')

                for office in entry.get('offices', []):
                    note = office['name']
                    for key, value in office.items():
                        if not value or key in ["name", "type"]:
                            continue
                        m.add_contact_detail(type=key, value=value, note=note)

                self.save_object(m)

            for session in entry.get('old_roles', []):
                roles = entry['old_roles'][session]
                meta = get_metadata(obj_to_jid(who))
                term = None

                for role in roles:
                    term = role['term']
                    to = None
                    for to in meta['terms']:
                        if term == to['name']:
                            term = to
                            break
                    else:
                        raise Exception("No term found?")

                    start_year = term['start_year']
                    end_year = term['end_year']

                    if 'committee' in role:
                        cid = role.get('committee_id')
                        if cid:
                            jid = _hot_cache.get(cid)
                            if jid:
                                m = Membership(who._id, jid,
                                               role='member',
                                               start_date=str(start_year),
                                               end_date=str(end_year),
                                               chamber=role['chamber'])
                                m.add_extra('term', term['name'])

                                if "position" in role:
                                    m.role = role['position']

                                self.save_object(m)

                                if m.role != 'member':
                                    # In addition to being the (chair|vice-chair),
                                    # they should also be noted as a member.
                                    m = Membership(who._id,
                                                   jid,
                                                   role='member',
                                                   start_date=str(start_year),
                                                   end_date=str(end_year),
                                                   chamber=role['chamber'])
                                    m.add_extra('term', term['name'])
                                    self.save_object(m)

                    if 'district' in role:
                        oid = "{state}-{chamber}".format(**role)
                        leg_ocdid = _hot_cache.get(oid)

                        geo_id = self.get_slug(
                            entry['state'], role['chamber'], role['district']
                        )

                        if leg_ocdid:
                            m = Membership(who._id, leg_ocdid,
                                           start_date=str(start_year),
                                           end_date=str(end_year),
                                           division_id=geo_id,
                                           post_id=role['district'],
                                           chamber=role['chamber'])
                            m.add_extra('term', term['name'])
                            self.save_object(m)

    def get_slug(self, state, chamber, district):
        path = "%s/%s.csv" % (self.mapping_dir, state)
        for row in csv.DictReader(open(path, 'r').readlines()):
            if (row['chamber'] == chamber and
                    row['name'].lower().strip() == district.lower().strip() and
                    row['abbr'] == state):

                _, slug = row['boundary_id'].rsplit("/", 1)
                _, slug = slug.split("-", 1)

                osid = "%s-%s-%s" % (state, chamber, slug)
                return self.osid_to_geo_id[osid]['id']

        if "-" in district or "," in district:
            # Hack for MA.
            if "," in district:
                nd, _ = district.rsplit(",", 1)
            else:
                nd, _ = district.rsplit("-", 1)

            return self.get_slug(state, chamber, nd.strip())

        print("")
        print("Note: Not finding geo-id for %s %s %s" % (
            state, chamber, district))
        print("")

    def migrate_bills(self, state):
        spec = {}
        if state:
            spec['state'] = state

        bills = self.billy_db.bills.find(spec, timeout=False)
        for bill in bills:

            ocdid = _hot_cache.get('{state}-{chamber}'.format(**bill))
            org_name = name_cache.get(ocdid)
            if not org_name or not ocdid:
                raise Exception("""Can't look up chamber this legislative
                                instrument was introduced into.""")

            b = Bill(name=bill['bill_id'],
                     organization=org_name,
                     organization_id=ocdid,
                     session=bill['session'],
                     title=bill['title'],
                     chamber=bill['chamber'],
                     type=bill['type'],
                     created_at=bill['created_at'],
                     updated_at=bill['updated_at'])

            b.identifiers = [{'scheme': 'openstates',
                             'identifier': bill['_id']}]
            b._openstates_id = bill['_id']

            blacklist = ["bill_id", "session", "title", "chamber", "type",
                         "created_at", "updated_at", "sponsors", "actions",
                         "versions", "sources", "state", "action_dates",
                         "documents", "level", "country", "alternate_titles",
                         "subjects", "_id", "type", "_type", "_term", "_all_ids",
                         "summary", "_current_term", "_current_session"]

            for key, value in bill.items():
                if key in blacklist or not value:  # or key.startswith("_"):
                    continue
                b.extras[key] = value

            for subject in bill.get('subjects', []):
                b.add_subject(subject)

            if 'summary' in bill and bill['summary']:
                # OpenStates only has one at most. let's just convert it
                # blind.
                b.add_summary('summary', bill['summary'])

            if 'alternate_titles' in bill:
                b.other_titles = [{"title": x, "note": None}
                                  for x in bill['alternate_titles']]

            for source in bill['sources']:
                b.add_source(source['url'], note='OpenStates source')

            for document in bill['documents']:
                b.add_document_link(
                    mimetype=document.get('mimetype'),
                    name=document['name'],
                    url=document['url'],
                    document_id=document['doc_id'],
                    on_duplicate='ignore')  # Old docs are bad about this

            b.add_extra('action_dates', bill['action_dates'])

            for version in bill['versions']:
                kwargs = {}
                mime = version.get("mimetype", version.get("+mimetype"))
                if mime:
                    kwargs['mimetype'] = mime

                b.add_version_link(name=version['name'],
                                   url=version['url'],
                                   on_duplicate='ignore',
                                   document_id=version['doc_id'],
                                   # ^^^ Some old OS entries are not so
                                   # good about this.
                                   **kwargs)

            for subject in bill.get('subjects', []):
                b.add_subject(subject)

            for action in bill['actions']:
                related_entities = None
                related = action.get("related_entities")
                if related:
                    related_entities = []
                    for rentry in related:
                        type_ = {
                            "committee": "organizations",
                            "legislator": "people"
                        }[rentry['type']]

                        nid = rentry['id'] = lookup_entry_id(type_, rentry['id'])
                        if nid:
                            if "ocd" not in nid:
                                raise Exception("Non-OCD id")

                        rentry['_type'] = {
                            "committee": "organization",
                            "legislator": "person"
                        }[rentry.pop('type')]

                        related_entities.append(rentry)

                when = dt.datetime.strftime(action['date'], "%Y-%m-%d")

                translate = {"bill:introduced": "introduced",
                             "bill:reading:1": "reading:1",
                             "bill:reading:2": "reading:2",
                             "bill:reading:3": "reading:3"}

                type_ = [translate.get(x, None) for x in action['type']]

                b.add_action(description=action['action'],
                             actor=action['actor'],
                             date=when,
                             type=filter(lambda x: x is not None, type_),
                             related_entities=related_entities)

            for sponsor in bill['sponsors']:
                type_ = 'people'
                sponsor_id = sponsor.get('leg_id', None)

                if sponsor_id is None:
                    type_ = 'organizations'
                    sponsor_id = sponsor.get('committee_id', None)

                kwargs = {}
                if sponsor_id:
                    objid = lookup_entry_id(type_, sponsor_id)

                    if objid is not None:
                        kwargs['entity_id'] = objid

                etype = {"people": "person",
                         "organizations": "organization"}[type_]

                #if sponsor.get('official_type'):
                #    kwargs['official_type'] = sponsor['official_type']
                # Not needed???

                b.add_sponsor(
                    name=sponsor['name'],
                    sponsorship_type=sponsor['type'],
                    entity_type=etype,
                    primary=sponsor['type'] == 'primary',
                    chamber=sponsor.get('chamber', None),
                    **kwargs)

            self.save_object(b)

    def migrate_votes(self, state):
        spec = {}
        if state:
            spec['state'] = state

        for entry in self.billy_db.votes.find(spec, timeout=False):
            #def __init__(self, session, date, type, passed,
            #             yes_count, no_count, other_count=0,
            #             chamber=None, **kwargs):

            when = dt.datetime.strftime(entry['date'], "%Y-%m-%d")
            if entry.get('type') is None:
                continue

            org_name = entry.get('committee')
            if org_name:
                # There's a committee tied to this vote.
                org_id = entry.get('committee_id')
                ocdid = _hot_cache.get(org_id)
            else:
                # OK. We don't have a committee. We should have a chamber.
                # Let's fallback on the COW.
                ocdid = _hot_cache.get('{state}-{chamber}'.format(**entry))
                org_name = name_cache.get(ocdid)
                if ocdid is None or org_name is None:
                    if entry['chamber'] == 'joint':
                        print("""
    XXX: Vote is marked as joint, but has no committee.
    This is likely (totally is) a bug. Please fix the scraper. We'll pass on it in the meantime.
    VoteID: %s" % (entry['_id'])""")
                        continue

                    raise Exception("""Can't look up the legislature? Something
                                     went wrong internally. The cache might be
                                     wrong for some reason. Look into this.""")

            v = Vote(
                organization=org_name,
                organization_id=ocdid,
                motion=entry['motion'],
                session=entry['session'],
                date=when,
                type=entry['type'],
                passed=entry['passed'],
                yes_count=entry['yes_count'],
                no_count=entry['no_count'],
                other_count=entry['other_count'],
                chamber=entry['chamber'],
                #created_at=entry['created_at'],
                #updated_at=entry['updated_at'],
            )
            v.identifiers = [{'scheme': 'openstates',
                             'identifier': entry['_id']}]
            v._openstates_id = entry['_id']

            for source in entry['sources']:
                v.add_source(**source)

            if v.sources == []:
                continue  # emit warning

            for vtype in ['yes', 'no', 'other']:
                for voter in entry["%s_votes" % (vtype)]:
                    id = voter.get('leg_id', None)
                    hcid = _hot_cache.get(id, None)
                    getattr(v, vtype)(name=voter['name'], id=hcid)

            hcid = _hot_cache.get(entry['bill_id'], None)
            bid = hcid

            # The bill_id coming off the entry is an ugly OpenStates bigID.
            # We need to actually get the bill and use that for this. This
            # results in kinda a silly slowdown. Perhaps we can cache titles.

            bill_id = entry['bill_id']  # as fallback.
            if bid:
                if bid in name_cache:
                    bill_id = name_cache[bid]

                # The following is the old code path. However, if the bill didn't
                # get converted, we won't have the object in the database anyway,
                # so, we'll make a memory/network tradeoff here.
                #                       -- PRT
                #
                # dbill = db.bills.find_one({"_id": bid})
                # if dbill:
                #     bill_id = dbill['name']

            v.set_bill(what=bill_id, id=bid, chamber=entry['chamber'])

            self.save_object(v)

    def migrate_events(self, state):
        spec = {}
        if state:
            spec['state'] = state

        for entry in self.billy_db.events.find(spec, timeout=False):

            e = Event(
                name=entry['description'],
                when=entry['when'],
                location=entry['location'],
                session=entry['session'],
                updated_at=entry['updated_at'],
                created_at=entry['created_at'],
                type=entry['type'],
            )
            e.identifiers = [{'scheme': 'openstates',
                             'identifier': entry['_id']}]
            e._openstates_id = entry['_id']
            if entry.get('+location_url'):
                e.add_location_url(entry['+location_url'])

            link = entry.get('link', entry.get("+link"))
            if link:
                e.add_link(link, 'link')

            blacklist = ["description", "when", "location", "session",
                         "updated_at", "created_at", "end", "sources",
                         "documents", "related_bills", "state", "+link",
                         "link", "level", "participants", "country",
                         "_all_ids", "type"]

            e.status = entry.get('status')
            typos = {
                "canceled": "cancelled"
            }
            if e.status in typos:
                e.status = typos[e.status]

            for key, value in entry.items():
                if key in blacklist or not value or key.startswith("_"):
                    continue
                e.extras[key] = value

            if entry.get('end'):
                end = entry['end']
                try:
                    end = dt.datetime.fromtimestamp(end)
                except TypeError:
                    pass

                e.end = end

            for source in entry['sources']:
                e.add_source(url=source['url'])

            if e.sources == []:
                continue  # XXX: print warning

            for document in entry.get('documents', []):
                e.add_document(name=document.get('name'),
                               document_id=document.get('doc_id'),
                               url=document['url'],
                               mimetype=document.get(
                                   "mimetype", document.get(
                                       "+mimetype",
                                       "application/octet-stream")))
                # Try to add the mimetype. If it fails, fall back to a generic
                # undeclared application/octet-stream.

            agenda = None
            for bill in entry.get('related_bills', []):
                if agenda is None:
                    agenda = e.add_agenda_item(
                        description="Bills up for Consideration"
                    )

                hcid = _hot_cache.get(bill.get('id', None), None)
                bid = bill['bill_id']
                if bid is None:
                    continue

                agenda.add_bill(bill=bid, id=hcid)

            for who in entry.get('participants', []):
                participant_type = who.get('participant_type', 'committee')
                # I've gone through the backlog of OpenStates data, they are
                # all committees of some sort.

                who_chamber = who.get('chamber')
                if who_chamber is None:
                    for chamber in ["_chamber", "+chamber"]:
                        f = who.get(chamber)
                        if f:
                            who_chamber = f
                            break

                if who_chamber is None:
                    # Freak of nature ...
                    continue

                hcid = _hot_cache.get(who.get('id', None), None)

                e.add_participant(
                    name=who['participant'],
                    type={
                        "committee": "organization",
                        "legislator": "person",
                        "person": "person",
                    }[participant_type],
                    id=hcid,
                    note=who['type'],
                    chamber=who_chamber)

            self.save_object(e)

    def save_objects(self, payload):
        for entry in payload:

            if self.validate:
                entry.validate()

            what = type_tables[type(entry)]
            table = getattr(db, what)

            _id = None
            try:
                _id = entry._id
            except AttributeError:
                pass

            ocd_id = ocd_namer(entry)
            if ocd_id:
                if _id and not is_ocd_id(_id):
                    _id = None

                if _id is None:
                    entry._id = ocd_id

            if what != 'people' and getattr(entry, '_openstates_id', None):
                jid = obj_to_jid(entry)
                entry.jurisdiction_id = jid

            entry.add_meta_software('openstates')
            eo = entry.as_dict()
            mongo_id = table.save(eo)

            if _id is None and ocd_id is None:
                entry._id = mongo_id

            if hasattr(entry, "_openstates_id"):
                _hot_cache[entry._openstates_id] = entry._id
                _cache_touched[entry._openstates_id] = True

            if hasattr(entry, "name"):
                name_cache[entry._id] = entry.name

            if self.quiet:
                sys.stdout.write(entry._type[0])
                sys.stdout.flush()

    def save_object(self, payload):
        return self.save_objects([payload])



from __future__ import print_function
import shutil
import importlib
import sys
import traceback
from os.path import join, abspath, dirname

from .base import BaseCommand


class UpdateError(Exception):
    pass


class Command(BaseCommand):
    name = 'init'
    help = 'start a new pupa scraper'

    def add_args(self):
        # what to scrape
        self.add_argument('module', type=str, help='name of the new scraper module')

        # debugging
        self.add_argument('--debug', nargs='?', const='pdb', default=None,
                          help='drop into pdb (or set =ipdb =pudb)')

    def enable_debug(self, debug):
        # turn debug on
        if debug:
            _debugger = importlib.import_module(debug)

            # turn on PDB-on-error mode
            # stolen from http://stackoverflow.com/questions/1237379/
            # if this causes problems in interactive mode check that page
            def _tb_info(type, value, tb):
                traceback.print_exception(type, value, tb)
                _debugger.pm()
            sys.excepthook = _tb_info

    def handle(self, args):
        self.enable_debug(args.debug)
        pupa_dir = dirname(abspath(__file__))
        example_dir = join(pupa_dir, '../../../example')
        ignore = shutil.ignore_patterns('*.pyc', '__pycache__')
        try:
            shutil.copytree(example_dir, args.module, ignore=ignore)
        except OSError as exc:
            if exc.errno:
                print('Error: the folder %r already exists.' % args.module)



contact_details = {
    "description": "Contact information for this entity.",
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "type": {"type": "string",
                     "description": "type of contact (e.g. phone, email, address)"},
            "value": {"type": "string",
                      "description": "actual phone number/email address/etc.",
                     },
            "note": {"type": ["string", "null"],
                     "description": "for grouping data by location/etc."
                    },
            "label": {"type": ["string", "null"],
                      "description": "human readable label",
                     },
        }
    }
}

identifiers = {
    "items": {
        "properties": {
            "identifier": {"type": "string",
                           "description": "The 3rd-party identifier, such as OKL0001000."},
            "scheme": {"type": ["string", "null"],
                       "description": "What service this identifier is used by."
                      },
        }
    },
    "type": "array",
    "description": "IDs other than the primary ID that the object may be known by."
}

other_names = {
    "description": "Alternate or former names for this object.",
    "items": {
        "properties": {
            "name": {"type": "string",
                     "description": "An alternate name this object is sometimes known by.",
                    },
            "start_date": {
                "type": ["string", "null"],
                "pattern": "^[0-9]{4}(-[0-9]{2}){0,2}$",
                "description": ("The date at which this name became valid."
                                "(null if unknown/valid indefinitely)"),
            },
            "end_date": {
                "type": ["string", "null"],
                "pattern": "^[0-9]{4}(-[0-9]{2}){0,2}$",
                "description": ("The date at which this name was no longer valid. "
                                "(null if still valid/valid indefinitely)"),
            },
            "note": {"type": ["string", "null"],
                     "description": ("An optional note describing where this alternate name came "
                                     "from or its relationship to the entity."),
                    }
        },
        "type": "object",
    },
    "type": "array"
}


links = {
    "description": "URLs for documents about the person",
    "items": {
        "properties": {
            "note": {
                "description": "A note, e.g. 'Wikipedia page'",
                "type": ["string", "null"],
            },
            "url": {
                "description": "A URL for a document about the person",
                "format": "uri",
                "type": "string"
            }
        },
        "type": "object"
    },
    "type": "array"
}


sources = {
    "description": "URLs for sources relating to the object",
    "items": {
        "properties": {
            "url": {
                "type": "string",
                "description": "URL of resource used to collect information",
            },
            "note": {
                "type": ["null", "string"],
                "description": "note about what information this URL was used for",
            }
        },
        "type": "object"
    },
    "minItems": 1,
    "type": "array"
}



import argparse
import logging
import importlib

logging.basicConfig()
logger = logging.getLogger('pupa')

COMMAND_MODULES = (
    'pupa.cli.commands.update',
    'pupa.cli.commands.ensure_indexes',
    'pupa.cli.commands.init',
    'pupa.cli.commands.import_billy',
    'pupa.cli.commands.dump',
)


def main():
    parser = argparse.ArgumentParser('pupa', description='pupa CLI')
    subparsers = parser.add_subparsers(dest='subcommand')

    subcommands = {}
    for mod in COMMAND_MODULES:
        try:
            cmd = importlib.import_module(mod).Command(subparsers)
            subcommands[cmd.name] = cmd
        except ImportError as e:
            logger.warning('error "%s" prevented loading of %s module',
                           e, mod)

    # process args
    args = parser.parse_args()
    subcommands[args.subcommand].handle(args)


if __name__ == '__main__':
    main()



from .base import BaseModel
from .bill import Bill
from .schemas.vote import schema


class Vote(BaseModel):
    """
    """
    _type = 'vote'
    _schema = schema
    _collection = 'votes'
    __slots__ = ("session", "chamber", "date", "motion", "type", "passed",
                 "organization", "organization_id", "bill", "vote_counts",
                 "roll_call", "sources", '_openstates_id', 'jurisdiction_id',
                 'identifiers')

    def __init__(self, organization, session, date, motion, type, passed,
                 yes_count, no_count, other_count=0, organization_id=None,
                 chamber=None, **kwargs):

        super(Vote, self).__init__()

        if not isinstance(type, list):
            type = [type]

        self.organization = organization
        self.organization_id = organization_id
        self.session = session
        self.date = date
        self.motion = motion
        self.type = type
        self.passed = passed
        self.chamber = chamber
        self.roll_call = []

        self.vote_counts = [
            {"vote_type": "yes", "count": yes_count},
            {"vote_type": "no", "count": no_count},
            {"vote_type": "other", "count": other_count},
        ]

        for k, v in kwargs.items():
            setattr(self, k, v)

    def __str__(self):
        return u'{0} - {1} - {2}'.format(self.session, self.date, self.motion)

    __unicode__ = __str__

    def set_bill(self, what, id=None, chamber=None):
        """
        COMPLEX BEHAVIOR AHOY:

            If `what` is a `Bill` object, this will correctly link without
            needing *anything* else. Adding `id` or `chamber` will result in
            a `TypeError` being raised.

            If `what` is not a `Bill` object, this will set `bill.name` to this
            value, and apply the rest of the keywords to the `bill` attribute.
        """
        if isinstance(what, Bill):
            if id or chamber:
                raise TypeError("set_bill takes no arguments "
                                "when using a `Bill` object")
            return self._set_bill_obj(what)
        return self._set_bill_raw(name=what, id=id, chamber=chamber)

    def _set_bill_raw(self, name, id=None, chamber=None):
        self.bill = {
            "id": id,
            "name": name,
            "chamber": chamber
        }

    def _set_bill_obj(self, bill):
        self._set_bill_raw(name=bill.name, id=bill._id)
        if hasattr(bill, 'chamber') and bill.chamber:
            self.bill['chamber'] = bill.chamber

    def vote(self, name, how, id=None):
        self.roll_call.append({
            "vote_type": how,
            "person": {
                "name": name,
                "id": id,
            }
        })

    def yes(self, name, id=None, **kwargs):
        return self.vote(name, 'yes', id=id, **kwargs)

    def no(self, name, id=None, **kwargs):
        return self.vote(name, 'no', id=id, **kwargs)

    def other(self, name, id=None, **kwargs):
        return self.vote(name, 'other', id=id, **kwargs)



"""
    Schema for event objects.
"""

from .common import sources

media_schema = {
    "description": ("This \"special\" schema is used in two places in the Event"
                    " schema, on the top level and inside the agenda item. This is an"
                    " optional component that may be omited entirely from a document."),
    "items": {
        "properties": {
            "name": {
                "type": "string",
                "description": ('name of the media link, such as "Recording of'
                                ' the meeting" or "Discussion of construction'
                                ' near the watershed"'),
            },

            "type": {
                "type": "string",
                "description": ('type of the set of recordings, such as'
                                ' "recording" or "testimony".'),
            },

            "date": {
                "pattern": "^[0-9]{4}(-[0-9]{2}){0,2}$",
                "type": ["string", "null"],
                "description": "date of the recording.",
            },

            "offset": {
                "type": ["number", "null"],
                "description": ("Offset where the related part starts. This is"
                                " optional and may be ommited entirely."),
            },

            "links": {
                "description": ("List of links to the same media item, each"
                                " with a different MIME type."),
                "items": {
                    "properties": {
                        "mimetype": {
                            "description": ("Mimetype of the media, such"
                                            " as video/mp4 or audio/webm"),
                            "type": ["string", "null"]
                        },

                        "url": {
                            "type": "string",
                            "description": "URL where this media may be accessed",
                        },

                    },
                    "type": "object"
                },
                "type": "array"
            },
        },
        "type": "object"
    },
    "type": "array"
}

schema = {
    "description": "event data",

    "_order": (
        ('Basics', ('_type', 'name', 'description', 'when', 'end', 'status', 'location')),
        ('Linked Entities', ('media', 'links', 'participants', 'agenda', 'documents',)),
        ('Common Fields', ['updated_at', 'created_at', 'sources']),
    ),

    "properties": {
        "_type": {
            "enum": ["event"],
            "type": "string",
            "description": ("All events must have a _type field set to one of"
                            " the entries in the enum below."),
        },

        "name": {
            "type": "string",
            "description": ('A simple name of the event, such as "Fiscal'
                            ' subcommittee hearing on pudding cups"')
        },

        "updated_at": {
            "type": ["string", "datetime"],
            "required": False,
            "description": "the time that this object was last updated.",
        },

        "created_at": {
            "type": ["string", "datetime"],
            "required": False,
            "description": "the time that this object was first created.",
        },

        "description": {
            "type": ["string", "null"],
            "description": ('A longer description describing the event. As an'
                            ' example, "Topics for discussion include this that'
                            ' and the other thing. In addition, lunch will be'
                            ' served".'),
        },

        "when": {
            "type": ["datetime"],
            "description": ("Starting date / time of the event. This should be"
                            " fully timezone qualified."),
        },

        "end": {
            "type": ["datetime", "null"],
            "description": ("Ending date / time of the event. This should"
                            " be fully timezone qualified."),
        },


        "status": {
            "type": ["string", "null"],
            "enum": ["cancelled", "tentative", "confirmed", "passed"],
            "description": ("String that denotes the status of the meeting."
                            " This is useful for showing the meeting is cancelled"
                            " in a machine-readable way."),
        },

        "location": {
            "description": "Where the event will take place.",
            "type": "object",
            "properties": {

                "name": {
                    "type": "string",
                    "description": ('name of the location, such as "City Hall,'
                                    ' Boston, MA, USA", or "Room E201, Dolan'
                                    ' Science Center, 20700 North Park Blvd'
                                    ' University Heights Ohio, 44118"'),
                },

                "note": {
                    "type": ["string", "null"],
                    "description": ('human readable notes regarding the location,'
                                    ' something like "The meeting will take place'
                                    ' at the Minority Whip\'s desk on the floor"')
                },

                "url": {
                    "required": False,
                    "type": "string",
                    "description": "URL of the location, if applicable.",
                },

                "coordinates": {
                    "description": ('coordinates where this event will take'
                                    ' place. If the location hasn\'t (or isn\'t)'
                                    ' geolocated or geocodable, than this should'
                                    ' be set to null.'),
                    "type": ["object", "null"],
                    "properties": {
                        "latitude": {
                            "type": "string",
                            "description": "latitude of the location, if any",
                        },

                        "longitude": {
                            "type": "string",
                            "description": "longitude of the location, if any",
                        }
                    }
                },
            },
        },

        "media": media_schema,

        "documents": {
            "description": ("Links to related documents for the event. Usually,"
                            " this includes things like pre-written testimony,"
                            " spreadsheets or a slide deck that a presenter will"
                            " use."),
            "items": {
                "properties": {
                    "name": {
                        "type": "string",
                        "description": ('name of the document. Something like'
                                        ' "Fiscal Report" or "John Smith\'s'
                                        ' Slides".'),
                    },

                    "url": {
                        "type": "string",
                        "description": "URL where the content may be found.",
                    },

                    "mimetype": {
                        "type": "string",
                        "description": "Mimetype of the document.",
                    },
                },
                "type": "object"
            },
            "type": "array"
        },

        "links": {
            "description": ("Links related to the event that are not documents"
                            " or items in the Agenda. This is filled with helpful"
                            " links for the event, such as a committee's homepage,"
                            " reference material or links to learn more about subjects"
                            " related to the event."),
            "items": {
                "properties": {

                    "note": {
                        "description": ('Human-readable name of the link. Something'
                                        ' like "Historical precedent for popsicle procurement"'),
                        "type": ["string", "null"]
                    },

                    "url": {
                        "description": "A URL for a link about the event",
                        "format": "uri",
                        "type": "string"
                    }
                },
                "type": "object"
            },
            "type": "array"
        },

        "participants": {
            "description": ("List of participants in the event. This includes"
                            " committees invited, legislators chairing the event"
                            " or people who are attending."),
            "items": {
                "properties": {
                    "chamber": {
                        "type": ["string", "null"],
                        "description": ("Optional field storing the chamber of"
                                        " the related participant."),
                    },

                    "name": {
                        "type": "string",
                        "description": "Human readable name of the entitity.",
                    },

                    "id": {
                        "type": ["string", "null"],
                        "description": "ID of the participant",
                    },

                    "type": {
                        "enum": ["organization", "person"],
                        "type": "string",
                        "description": ("What type of entity is this? `person`"
                                        " may be used if the person is not a Legislator,"
                                        " butattending the event, such as an"
                                        " invited speaker or one who is offering"
                                        " testimony."),
                    },

                    "note": {
                        "type": "string",
                        "description": ("Note regarding the relationship, such"
                                        " as `chair` for the chair of a meeting."),
                    },

                },
                "type": "object"
            },
            "type": "array"
        },

        "agenda": {
            "description": ("Agenda of the event, if any. This contains information"
                            " about the meeting's agenda, such as bills to"
                            " discuss or people to present."),
            "items": {
                "properties": {
                    "description": {
                        "type": "string",

                        "description": ("Human-readable string that represents this"
                                        " agenda item. A good example would be something like"
                                        " The Committee will consider SB 2339, HB 100"),
                    },

                    "order": {
                        "type": ["string", "null"],
                        "description": ("order of this item, useful for re-creating"
                                        " meeting minutes. This may be ommited entirely."
                                        " It may also optionally contains \"dots\""
                                        " to denote nested agenda items, such as \"1.1.2.1\""
                                        " or \"2\", which may go on as needed."),
                    },

                    "subjects": {
                        "description": ("List of related topics of this agenda"
                                        " item relates to."),
                        "items": {"type": "string"},
                        "type": "array"
                    },

                    "media": media_schema,

                    "notes": {
                        "description": ("List of notes taken during this agenda"
                                        " item, may be used to construct meeting minutes."),
                        "items": {
                            "properties": {
                                "description": {
                                    "type": "string",
                                    "description": ("simple string containing the"
                                                    " content of the note."),
                                },
                            },
                            "type": "object"
                        },
                        "type": "array"
                    },

                    "related_entities": {
                        "description": ("Entities that relate to this agenda"
                                        " item, such as presenters, legislative"
                                        " instruments, or committees."),
                        "items": {
                            "properties": {
                                "type": {
                                    "type": "string",
                                    "description": ("type of the related object, like"
                                                    " `bill` or `organization`."),
                                },

                                "id": {
                                    "type": ["string", "null"],
                                    "description": "ID of the related entity",
                                },

                                "name": {
                                    "type": "string",
                                    "description": ("human readable string"
                                                    " representing the entity,"
                                                    " such as `John Q. Smith`."),
                                },

                                "note": {
                                    "type": ["string", "null"],
                                    "description": ("human readable string (if any) noting"
                                                    " the relationship between the entity and"
                                                    " the agenda item, such as \"Jeff"
                                                    " will be presenting on the effects"
                                                    " of too much cookie dough\""),
                                },
                            },
                            "type": "object",
                        },
                        "minItems": 0,
                        "type": "array",
                    },
                },
                "type": "object"
            },
            "minItems": 0,
            "type": "array"
        },
        "sources": sources,
    },
    "type": "object"
}



from .common import links, contact_details, identifiers, other_names, sources

schema = {
    "$schema": "http://json-schema.org/draft-03/schema#",
    "description": "A real person, alive or dead",
    "id": "http://popoloproject.com/schemas/person.json#",
    "_order": (
        ('Basics', ('name', 'image', 'contact_details', 'links')),
        ('Extended Details', ('gender', 'summary', 'biography', 'birth_date', 'death_date')),
        ('Alternate Names/Identifiers', ('identifiers', 'other_names')),
        ('Common Fields', ('updated_at', 'created_at', 'sources')),
    ),
    "properties": {
        "updated_at": {
            "description": "The time at which the resource was last modified",
            "type": ["string", "datetime", "null"],
        },
        "created_at": {
            "description": "The time at which the resource was created",
            "type": ["string", "datetime", "null"],
        },
        "biography": {
            "description": "An extended account of a person's life",
            "type": ["string", "null"],
        },
        "birth_date": {
            "description": "A date of birth",
            "pattern": "^[0-9]{4}(-[0-9]{2}){0,2}$",
            "type": ["string", "null"],
        },
        "death_date": {
            "description": "A date of death",
            "pattern": "^[0-9]{4}(-[0-9]{2}){0,2}$",
            "type": ["string", "null"],
        },
        "gender": {
            "description": "A gender",
            "type": ["string", "null"],
        },
        "image": {
            "description": "A URL of a head shot",
            "format": "uri",
            "type": ["string", "null"],
        },
        "name": {
            "description": "A person's preferred full name",
            "type": "string",
        },
        "contact_details": contact_details,
        "other_names": other_names,
        "links": links,
        "identifiers": identifiers,
        "summary": {
            "description": "A one-line account of a person's life",
            "type": ["string", "null"],
        },
        "national_identity": {
            "description": "A national identity",
            "type": ["string", "null"],
        },
        "sources": sources,
    },
    "title": "Person",
    "type": "object"
}



import os
import json
import logging
import datetime
from collections import defaultdict

import scrapelib

from pupa import utils
from pupa.models.membership import Membership
from pupa.core import settings


class ScrapeError(Exception):
    pass


class Scraper(scrapelib.Scraper):
    """ Base class for all scrapers """

    class ContinueScraping(Exception):
        """ indicate that scraping should continue without saving an object """
        pass

    def __init__(self, jurisdiction, session, output_dir, cache_dir=None,
                 strict_validation=True, fastmode=False):

        super(Scraper, self).__init__()

        self.skipped = 0

        # set options
        self.jurisdiction = jurisdiction
        self.session = session

        # scrapelib setup
        self.timeout = settings.SCRAPELIB_TIMEOUT
        self.requests_per_minute = settings.SCRAPELIB_RPM
        self.retry_attempts = settings.SCRAPELIB_RETRY_ATTEMPTS
        self.retry_wait_seconds = settings.SCRAPELIB_RETRY_WAIT_SECONDS
        self.follow_robots = False

        if fastmode:
            self.requests_per_minute = 0
            self.cache_write_only = False

        # directories
        self.output_dir = output_dir
        if cache_dir:
            self.cache_storage = scrapelib.FileCache(cache_dir)

        # validation
        self.strict_validation = strict_validation

        # 'type' -> {set of names}
        self.output_names = defaultdict(set)

        # logging convenience methods
        self.logger = logging.getLogger("pupa")
        self.info = self.logger.info
        self.debug = self.logger.debug
        self.warning = self.logger.warning
        self.error = self.logger.error
        self.critical = self.logger.critical

    def get_current_session(self):
        current_session = self.jurisdiction.terms[0]['sessions'][-1]
        return current_session

    def save_object(self, obj):
        if hasattr(obj, '_is_legislator'):
            membership = Membership(
                obj._id, 'jurisdiction:' + self.jurisdiction.jurisdiction_id,
                post_id=obj.post_id, chamber=obj.chamber,
                contact_details=obj._contact_details, role=obj._role)
            # remove placeholder _contact_details
            del obj._contact_details
            del obj._role
            obj._related.append(membership)

            # create a party membership
            if obj.party:
                membership = Membership(obj._id,
                                        'party:' + obj.party,
                                        role='member')
                obj._related.append(membership)

        filename = '{0}_{1}.json'.format(obj._type, obj._id)

        self.info('save %s %s as %s', obj._type, obj, filename)

        self.output_names[obj._type].add(filename)

        with open(os.path.join(self.output_dir, filename), 'w') as f:
            json.dump(obj.as_dict(), f, cls=utils.JSONEncoderPlus)

        # validate after writing, allows for inspection on failure
        try:
            obj.validate()
        except ValueError as ve:
            self.warning(ve)
            if self.strict_validation:
                raise ve

        # after saving and validating, save subordinate objects
        for obj in getattr(obj, '_related', []):
            self.save_object(obj)

    def _scrape(self, iterable, scrape_type):
        record = {'objects': defaultdict(int)}
        self.output_names = defaultdict(set)
        record['start'] = datetime.datetime.utcnow()
        for obj in iterable:
            if hasattr(obj, '__iter__'):
                for iterobj in obj:
                    self.save_object(iterobj)
            else:
                self.save_object(obj)
        record['end'] = datetime.datetime.utcnow()
        if not self.output_names:
            raise ScrapeError("no objects returned from {0} scrape".format(
                scrape_type))
        for _type, nameset in self.output_names.items():
            record['objects'][_type] += len(nameset)

        return {scrape_type: record}

    def scrape_people(self):
        return self._scrape(self.get_people(), 'people')

    def scrape_bills(self):
        return self._scrape(self.get_bills(), 'bills')

    def scrape_votes(self):
        return self._scrape(self.get_votes(), 'votes')

    def scrape_events(self):
        return self._scrape(self.get_events(), 'events')

    def scrape_speeches(self):
        return self._scrape(self.get_speeches(), 'speeches')

    def get_bills(self):
        for bill_id, extras in self.get_bill_ids():
            try:
                yield self.get_bill(bill_id, **extras)
            except self.ContinueScraping as exc:
                self.warning('skipping %s: %r', bill_id, exc)
                self.skipped += 1
                continue

    def get_people(self):
        raise NotImplementedError(self.__class__.__name__ + ' must provide a '
                                  'get_people() method')

    def get_bill_ids(self):
        raise NotImplementedError(self.__class__.__name__ + ' must provide a '
                                  'get_bill_ids() method or override '
                                  'get_bills()')

    def get_bill(self, bill_id, **kwargs):
        raise NotImplementedError(self.__class__.__name__ + ' must provide a '
                                  'get_bill() method or override '
                                  'get_bills()')

    def get_events(self):
        raise NotImplementedError(self.__class__.__name__ + ' must provide a '
                                  'get_events() method')

    def get_votes(self):
        raise NotImplementedError(self.__class__.__name__ + ' must provide a '
                                  'get_votes() method')

    def get_speeches(self):
        raise NotImplementedError(self.__class__.__name__ + ' must provide a '
                                  'get_speeches() method')



from collections import defaultdict


class CyclicGraphError(ValueError):
    """
    This exception is raised if the graph is Cyclic (or rather, when the
    sorting algorithm *knows* that the graph is Cyclic by hitting a snag
    in the top-sort)
    """
    pass


class Network(object):
    """
    This object (the `Network` object) handles keeping track of all the
    graph's nodes, and links between the nodes.

    The `Network' object is mostly used to topologically sort the nodes,
    to handle dependency resolution.
    """

    def __init__(self):
        self.nodes = []
        self.edges = defaultdict(set)

    def add_node(self, node):
        """ Add a node to the graph (with no edges) """
        self.nodes.append(node)

    def add_edge(self, fro, to):
        """
        Add an edge from node `fro` to node `to`. For instance, to say that
        `foo` depends on `bar`, you'd say::

            `network.add_edge('foo', 'bar')`
        """

        if fro not in self.nodes:
            self.add_node(fro)

        if to not in self.nodes:
            self.add_node(to)

        self.edges[fro].add(to)

    def leaf_nodes(self):
        """
        Return an interable of nodes with no edges pointing at them. This is
        helpful to find all nodes without dependencies.
        """
        deps = set([
            item for sublist in self.edges.values() for item in sublist
        ])  # Now contains all nodes that contain dependencies.
        return (x for x in self.nodes if x not in deps)  # Generator that
        # contains all nodes *without* any dependencies (leaf nodes)

    def prune_node(self, node, remove_backrefs=False):
        """
        remove node `node` from the network (including any edges that may
        have been pointing at `node`).
        """
        self.nodes = [x for x in self.nodes if x != node]
        if node in self.edges:
            # Remove add edges from this node if we're pruning it.
            self.edges.pop(node)

        for fro, connections in self.edges.items():
            # Remove any links to this node (if they exist)
            if node in self.edges[fro]:
                if remove_backrefs:
                    # If we should remove backrefs:
                    self.edges[fro].remove(node)
                else:
                    # Let's raise an Exception
                    raise ValueError("""Attempting to remove a node with
                                     backrefs. You may consider setting
                                     `remove_backrefs` to true.""")

    def sort(self):
        """
        Return an iterable of nodes, toplogically sorted to correctly import
        dependencies before leaf nodes.
        """
        while self.nodes != []:
            iterated = False
            for node in self.leaf_nodes():
                iterated = True
                self.prune_node(node)
                yield node
            if not iterated:
                raise CyclicGraphError("Sorting has found a cyclic graph.")



from six import text_type as str_type
from .base import BaseModel
from .utils import add_associated_link
from .schemas.bill import schema


def _cleanup_list(obj, default):
    if not obj:
        obj = default
    elif isinstance(obj, str_type):
        obj = [obj]
    elif not isinstance(obj, list):
        obj = list(obj)
    return obj


class Bill(BaseModel):
    """
    A single OpenCivic bill.
    """

    _type = 'bill'
    _schema = schema
    _collection = 'bills'
    __slots__ = ('actions', 'other_names', 'other_titles',
                 'related_bills', 'name', 'chamber', 'documents', 'session',
                 'sources', 'sponsors', 'summaries', 'subject', 'title',
                 '_openstates_id', 'type', 'versions', 'jurisdiction_id',
                 'organization', 'organization_id', 'identifiers')

    def __init__(self, name, session, title,
                 organization=None, type=None, **kwargs):
        super(Bill, self).__init__()

        self.name = name
        self.session = session
        self.title = title
        self.chamber = None
        self.type = _cleanup_list(type, ['bill'])
        self.organization = organization

        self.actions = []
        self.other_names = []
        self.other_titles = []
        self.documents = []
        self.related_bills = []
        self.sponsors = []
        self.subject = []
        self.summaries = []
        self.versions = []

        for k, v in kwargs.items():
            setattr(self, k, v)

    def add_action(self, description, actor, date,
                   type=None, related_entities=None):
        self.actions.append({
            "description": description,
            "actor": actor,
            "date": date,
            "type": _cleanup_list(type, []),
            "related_entities": related_entities or []  # validate
        })

    def add_related_bill(self, name, session, chamber, relation):
        self.related_bills.append({
            "name": name,
            "session": session,
            "chamber": chamber,
            "relation_type": relation  # validate
        })

    def add_sponsor(self, name, sponsorship_type,
                    entity_type, primary,
                    chamber=None, entity_id=None):
        ret = {
            "name": name,
            "sponsorship_type": sponsorship_type,
            "_type": entity_type,
            "primary": primary,
            "id": entity_id,
            "chamber": chamber,
        }

        self.sponsors.append(ret)

    def add_subject(self, subject):
        self.subject.append(subject)

    def add_document_link(self, name, url, date=None, type=None,
                          mimetype=None, on_duplicate='error',
                          document_id=None):
        return self._add_associated_link(
            collection='documents',
            name=name,
            url=url,
            date=date,
            type=type,
            mimetype=mimetype,
            on_duplicate=on_duplicate,
            document_id=document_id)

    def add_version_link(self, name, url, date=None, type=None,
                         mimetype=None, on_duplicate='error',
                         document_id=None):
        return self._add_associated_link(
            collection='versions',
            name=name,
            url=url,
            date=date,
            type=type,
            mimetype=mimetype,
            on_duplicate=on_duplicate,
            document_id=document_id)

    def add_summary(self, note, text):
        self.summaries.append({
            "note": note,
            "text": text,
        })

    def __str__(self):
        return self.name + ' in ' + self.session

    __unicode__ = __str__
    _add_associated_link = add_associated_link



from validictory.validator import SchemaValidator
import datetime


class DatetimeValidator(SchemaValidator):
    """ add 'datetime' type that verifies that it has a datetime instance """

    def validate_type_datetime(self, x):
        return isinstance(x, (datetime.date, datetime.datetime))


def add_associated_link(self, collection, name, url,
                        type, mimetype, on_duplicate, date=None, offset=None,
                        document_id=None):

    if on_duplicate not in ['error', 'ignore']:
        raise TypeError("Sorry; we accept either `error' or `ignore' for "
                        "on_duplicate behavior.")

    try:
        versions = getattr(self, collection)
    except AttributeError:
        versions = self[collection]

    ver = {'name': name, 'links': [], 'date': date, 'offset': offset,
           'type': type, 'document_id': document_id}

    seen_links = set()  # We iterate over everything anyway. Meh. Storing
    # as an instance var is actually non-trivial, since we abuse __slots__
    # for as_dict, and it's otherwise read-only or shared set() instance.

    matches = 0
    for version in versions:
        for link in version['links']:
            seen_links.add(link['url'])

        if False not in (ver.get(x) == version.get(x)
                         for x in ["name", "type", "date"]):
            matches = matches + 1
            ver = version

    if matches > 1:
        raise ValueError("Something went just very wrong internally")

    if url in seen_links:
        if on_duplicate == 'error':
            raise ValueError("Duplicate entry in `%s' - URL: `%s'" % (
                collection, url
            ))
        else:
            # This means we're in ignore mode. This situation right here
            # means we should *skip* adding this link silently and continue
            # on with our scrape. This should *ONLY* be used when there's
            # a site issue (Version 1 == Version 2 because of a bug) and
            # *NEVER* because "Current" happens to match "Version 3". Fix
            # that in the scraper, please.
            #  - PRT
            return None

    # OK. This is either new or old. Let's just go for it.
    ret = {'url': url, 'mimetype': mimetype}

    ver['links'].append(ret)

    if matches == 0:
        # in the event we've got a new entry; let's just insert it into
        # the versions on this object. Otherwise it'll get thrown in
        # automagically.
        versions.append(ver)

    return ver



import sys
import pprint
from os.path import join, abspath, dirname

import sh
import lxml.html
from libmproxy import proxy, flow

from pupa.scrape import Scraper
from pupa.models import Bill


class LegistarScraper(Scraper):
    '''
    Record yourself exporting the excel spreadsheet using mitmproxy.
    Save the mitm flows to a file called 'mitmdump.in' in the scraper folder.
    Good to go.

    Then subclass this like so:

    class BillScraper(LegstarScraper):
        url = 'https://chicago.legistar.com/Legislation.aspx'
        columns = (
                'bill_id', 'enactment_id', 'type', 'status',
                'created', 'action', 'title')

    Where columns are the columns from the excel sheet.
    '''

    def get_bill_ids(self):
        filename = sys.modules[self.__class__.__module__].__file__
        here = dirname(abspath(filename))
        infile = join(here, 'mitmdump.in')
        outfile = join(here, 'mitmdump.out')

        if self.requests_per_minute:
            sh.mitmdump(
                n=True,       # Don't start proxy server.
                c=infile,     # Client replay of saved flows.
                w=outfile)    # Output file.

        with open(outfile) as f:
            flows = list(flow.FlowReader(f).stream())
        workbook = flows[-1].response.content.decode('utf-8')

        doc = lxml.html.fromstring(workbook)

        for tr in doc.xpath('//tr')[1:]:
            vals = [td.text_content() for td in tr.xpath('td')]
            vals = [val.strip().replace(u'\xa0', '') for val in vals]
            keys = self.columns
            data = dict(zip(keys, vals))

            if not data['title']:
                continue
            _type = self.get_type(data.get('type').lower())
            if _type not in ['ordinance', 'resolution']:
                continue
            else:
                data['type'] = _type

            yield data.pop('bill_id'), data

    def get_type(self, _type):
        return _type

    def get_bill(self, bill_id, **kwargs):
        title = kwargs.pop('title')
        _type = kwargs.pop('type')
        bill = Bill(bill_id, self.session, title=title, type=_type)
        bill.add_source(self.url)
        return bill


class FlowTranslator(object):
    '''Read in a flow file, then spit out some requests code.
    '''
    def __init__(self, flow_filename):
        with open(flow_filename) as f:
            self.flows = list(flow.FlowReader(f).stream())
        self.flow_index = 0
        self.requests = []

    def __iter__(self):
        '''Yield data/code for each request in the sequences.
        '''
        while True:
            yield self.next()

    def next(self):
        try:
            flow = self.flows[self.flow_index]
        except IndexError:
            raise StopIteration
        requestdata = self._flow_to_data(flow)
        self.flow_index += 1
        return requestdata

    def previous_flow(self):
        index = self.flow_index - 1
        if 0 <= index:
            return self.flows[index]

    def _flow_to_data(self, flow):

        data = dict(
            url=flow.request.get_url(),
            method=flow.request.method)

        form = flow.request.get_form_urlencoded().items()
        if form is not None:
            params = dict(form)
            previous_flow = self.previous_flow()
            if previous_flow is not None:
                response = previous_flow.response
                html = response.get_decoded_content()
                doc = lxml.html.fromstring(html)
            for name in params:
                if not name.startswith('__'):
                    continue
                el = doc.xpath('//*[@name="%s"]' % name)[0]
                params[name] = el.attrib['value']

        data['params'] = params
        return data


if __name__ == '__main__':
    import sys
    import requests
    fn = sys.argv[1]
    cafile = '/home/thom/.mitmproxy/mitmproxy-ca-cert.p12'
    for data in FlowTranslator(fn):
        proxies = dict(https='localhost:8080')
        resp = requests.request(verify=cafile, proxies=proxies, **data)
        print resp
        del data['params']
        pprint.pprint(data)
    import pdb; pdb.set_trace()


import re
import os
import time
import json
import datetime
import subprocess

from bson import ObjectId


def makedirs(dname):
    if not os.path.isdir(dname):
        os.makedirs(dname)


# fixing bill ids
_bill_id_re = re.compile(r'([A-Z]*)\s*0*([-\d]+)')
_mi_bill_id_re = re.compile(r'(SJR|HJR)\s*([A-Z]+)')


def fix_bill_id(bill_id):
    bill_id = bill_id.replace('.', '')
    # special case for MI Joint Resolutions
    if _mi_bill_id_re.match(bill_id):
        return _mi_bill_id_re.sub(r'\1 \2', bill_id, 1).strip()
    return _bill_id_re.sub(r'\1 \2', bill_id, 1).strip()


class JSONEncoderPlus(json.JSONEncoder):
    """
    JSONEncoder that encodes datetime objects as Unix timestamps and mongo
    ObjectIds as strings.
    """
    def default(self, obj, **kwargs):
        if isinstance(obj, datetime.datetime):
            return time.mktime(obj.utctimetuple())
        elif isinstance(obj, datetime.date):
            return time.mktime(obj.timetuple())
        elif isinstance(obj, ObjectId):
            return str(obj)

        return super(JSONEncoderPlus, self).default(obj, **kwargs)


def convert_pdf(filename, type='xml'):
    commands = {'text': ['pdftotext', '-layout', filename, '-'],
                'text-nolayout': ['pdftotext', filename, '-'],
                'xml': ['pdftohtml', '-xml', '-stdout', filename],
                'html': ['pdftohtml', '-stdout', filename]}
    try:
        pipe = subprocess.Popen(commands[type], stdout=subprocess.PIPE,
                                close_fds=True).stdout
    except OSError as e:
        raise EnvironmentError("error running %s, missing executable? [%s]" %
                               ' '.join(commands[type]), e)
    data = pipe.read()
    pipe.close()
    return data



from pupa.core import db


def collection_by_jurisdiction(jurisdiction_id, collection, field):
    """ Find all things in a jurisdiction """
    ids = getattr(db, collection).find({
        "jurisdiction_id": jurisdiction_id,
    }).distinct(field)
    if None in ids:
        ids.remove(None)
    return ids


def people_by_jurisdiction(jurisdiction_id):
    """ Find all people by a jurisdiction """
    return collection_by_jurisdiction(
        jurisdiction_id,
        'memberships',
        'person_id'
    )


def orgs_by_jurisdiction(jurisdiction_id):
    """ Find all orgs by a jurisdiction """
    return collection_by_jurisdiction(jurisdiction_id, 'organizations', '_id')


def bills_by_jurisdiction(jurisdiction_id):
    """ Find all bills by a jurisdiction """
    return collection_by_jurisdiction(jurisdiction_id, 'bills', '_id')


def people_by_name(name, people_ids=None, **kwargs):
    """ Find all people by their name. Optional people_ids _id constraint """
    spec = kwargs.copy()
    spec.update({"$or": [
        { "name": name },
        { "other_names.name": name },
        { "identifiers.identifier": name },
    ]})
    if people_ids is not None:
        # This isn't a raw if conditional, since you could pass
        # an empty list.
        spec["_id"] = {"$in": people_ids}
    return db.people.find(spec)


def bills_by_name(name, bill_ids=None, **kwargs):
    """ Find all bills by their name. Optional bill_ids _id constraint """
    spec = kwargs.copy()
    spec.update({"$or": [
        { "name": name },
        { "bill_id": name },
        { "alternate_titles.title": name },
        { "alternate_bill_ids.bill_id": name },
    ]})
    if bill_ids is not None:
        # This isn't a raw if conditional, since you could pass
        # an empty list.
        spec["_id"] = {"$in": bill_ids}
    return db.people.find(spec)


def orgs_by_name(name, org_ids=None, **kwargs):
    """ Find all orgs their name. Optional org_ids _id constraint """
    spec = kwargs.copy()

    spec.update({"$or": [
        { "name": name },
        { "other_names.name": name },
        { "identifiers.identifier": name },
    ]})

    if org_ids is not None:
        # This isn't a raw if conditional, since you could pass
        # an empty list.
        spec["_id"] = {"$in": org_ids}
    return db.organizations.find(spec)


def people_by_jurisdiction_and_name(jurisdiction_id, name, **kwargs):
    people_ids = people_by_jurisdiction(jurisdiction_id)
    people = people_by_name(name, people_ids=people_ids, **kwargs)
    return people


def orgs_by_jurisdiction_and_name(jurisdiction_id, name, **kwargs):
    org_ids = orgs_by_jurisdiction(jurisdiction_id)
    orgs = orgs_by_name(name, org_ids=org_ids, **kwargs)
    return orgs


def bills_by_jurisdiction_and_name(jurisdiction_id, name, **kwargs):
    bill_ids = bills_by_jurisdiction(jurisdiction_id)
    bills = bills_by_name(name, bill_ids=bill_ids, **kwargs)
    return bills



from .common import links, contact_details

schema = {
    "$schema": "http://json-schema.org/draft-03/schema#",
    "description": "A relationship between a person and an organization",
    "id": "http://popoloproject.com/schemas/membership.json#",
    "properties": {
        "updated_at": {
            "description": "The time at which the resource was last modified",
            "type": ["string", "datetime", "null"],
        },
        "created_at": {
            "description": "The time at which the resource was created",
            "type": ["string", "datetime", "null"],
        },

        "organization_id": {
            "description": "The ID of the organization that is a party to the relationship",
            "type": "string",
        },
        # The membership's person may only be linked at import time, but objects
        # are validated at scrape time; therefore, the person_id is allowed to
        # have a type of "null" at scrape time.
        "person_id": {
            "description": "The ID of the person who is a party to the relationship",
            "type": ["string", "null"],
        },
        "post_id": {
            "description": "The ID of the post held by the person in the organization through this membership",
            "type": ["string", "null"],
        },
        "on_behalf_of_id": {
            "description": "The ID of the organization on whose behalf the person is a party to the relationship",
            "type": ["string", "null"],
        },
        "label": {
            "description": "A label describing the membership",
            "type": ["string", "null"],
        },
        "role": {
            "description": "The role that the person fulfills in the organization",
            "type": ["string", "null"],
        },
        "start_date": {
            "description": "The date on which the relationship began",
            "pattern": "^[0-9]{4}(-[0-9]{2}){0,2}$",
            "type": ["string", "null"],
        },
        "end_date": {
            "description": "The date on which the relationship ended",
            "pattern": "^[0-9]{4}(-[0-9]{2}){0,2}$",
            "type": ["string", "null"],
        },
        "contact_details": contact_details,
        "links": links,
    },
    "title": "Membership",
    "type": "object"
}



from ..organization import Organization
from validictory import ValidationError
from nose.tools import raises


def test_basic_invalid_organization():
    """ Make sure we can make an invalid orga """
    orga = Organization("name")
    orga.add_source(url='foo')
    orga.validate()

    orga.name = None

    @raises(ValidationError)
    def _():
        orga.validate()
    _()


def test_add_post():
    """ Test that we can hack posts in on the fly'"""
    orga = Organization("name")
    orga.add_source(url='foo')
    orga.validate()

    orga.add_post("Human Readable Name", "Chef")

    assert orga.posts[0]['role'] == "Chef"
    assert orga.posts[0]['label'] == "Human Readable Name"


    @raises(TypeError)
    def _():
        orga.add_identifier("id10t", foo="bar")
    _()

    orga.add_identifier("id10t")
    orga.add_identifier("l0l", scheme="kruft")

    assert orga.identifiers[-1]['scheme'] == "kruft"
    assert orga.identifiers[0]['identifier'] == "id10t"
    assert not hasattr(orga.identifiers[0], "scheme")


def test_add_contact():
    """ test we can add a contact detail to an orga """
    orga = Organization("name")
    orga.add_source(url='foo')
    orga.validate()

    orga.add_contact_detail(type='pigeon',
                            value='39328',
                            note='His name is Joe')

    orga.validate()



from .base import BaseImporter
from .utils import (people_by_jurisdiction, people_by_name,
                    people_by_jurisdiction_and_name)
from pupa.models import Membership
from pupa.core import db


def match_membership(membership, people=None):
    if membership.unmatched_legislator:
        if people is None:
            unmatched_name = membership['unmatched_legislator']['name']
            people = people_by_jurisdiction_and_name(
                unmatched_name,
                membership['jurisdiction_id'],
            )

        if people.count() == 1:
            person = people[0]
            #   + update membership with this person's details (if it's
            #     not there already)
            membership.person_id = person['_id']
            membership.unmatched_legislator = None
            return (True, membership)
    return (False, membership)


class MembershipImporter(BaseImporter):
    _type = 'membership'
    _model_class = Membership

    def __init__(self, jurisdiction_id, person_importer, org_importer):
        super(MembershipImporter, self).__init__(jurisdiction_id)
        self.person_importer = person_importer
        self.org_importer = org_importer

    def get_db_spec(self, membership):
        spec = {'organization_id': membership.organization_id,
                'person_id': membership.person_id,
                'role': membership.role,
                # if this is a historical role, only update historical roles
                'end_date': membership.end_date}

        if membership.post_id:
            # The reason for adding this in a conditional is that
            # we may not always know the post in the scraper, but we shouldn't
            # make a mess with membership that have been attached to a post
            # manually. If we know it in the scraper, it better be in the DB!
            spec['post_id'] = membership.post_id

        if membership.unmatched_legislator:
            spec['unmatched_legislator'] = membership.unmatched_legislator
            unmatched_name = membership.unmatched_legislator['name']

            people = people_by_jurisdiction_and_name(membership.jurisdiction_id, unmatched_name)

            matched, membership = match_membership(membership, people=people)

            if matched:
                # We were able to simply match. Let's roll with this.
                spec = self.get_db_spec(membership)

                # OK. Since we matched, let's also update existing
                # copies of this guy.

                #### This will find anyone who has our ID *OR* has a None.
                pspec = spec.copy()
                pspec['person_id'] = membership.person_id

                person = db.people.find_one({"_id": membership.person_id})
                if person is None:
                    # Um. OK. This membership is linked to a ghost-person.
                    print("Warning: %s is linked to an UNKNOWN PERSON" % membership['_id'])
                    # The proper behavior here isn't clear. Rather than let
                    # unclean data into the DB, I'm going to break the import
                    # process.
                    raise Exception(
                        """
                        Bad membership link that got matched. This is likely an
                        internal bug. Please look at this *CAREFULLY* to work
                        out proper behavior
                        """
                    )

                uspec = spec.copy()
                uspec['person_id'] = None
                uspec['unmatched_legislator.name'] = {
                    "$in": [
                        x['name'] for x in person['other_names']
                    ] + [person['name']]
                }
                spec = {"$or": [pspec, uspec]}
                return spec

            #### This will find anyone who has our name *OR* has a None.
            # (not the same code as above)
            pspec = spec.copy()
            pspec['person_id'] = {"$in": people.distinct("_id")}
            pspec.pop('unmatched_legislator')
            # Either we have this person, who's already matched (e.g.
            # everything already matches)

            uspec = spec.copy()
            uspec['person_id'] = None
            # Or we don't, and we need to find a membership with an
            # unmatched ID
            spec = {"$or": [pspec, uspec]}

        return spec

    def prepare_object_from_json(self, obj):
        org_json_id = obj['organization_id']

        if org_json_id.startswith("jurisdiction:"):
            _, jid = org_json_id.split(":", 1)
            org = self.org_importer._resolve_org_by_chamber(jid, obj['chamber'])
            if org is None:
                raise ValueError("Something went wrong with the Jurisidction "
                                 "import")
            obj['organization_id'] = org['_id']
        else:
            obj['organization_id'] = self.org_importer.resolve_json_id(org_json_id)

        person_json_id = obj['person_id']
        obj['person_id'] = self.person_importer.resolve_json_id(person_json_id)
        return obj



from .common import links, contact_details, identifiers, other_names, sources

CLASSIFICATIONS = ['legislature', 'party', 'committee', 'commission']

schema = {
    "properties": {
        "classification": {
            "description": "An organization category, e.g. committee",
            "type": ["string", "null"],
            "enum": CLASSIFICATIONS,
        },
        "dissolution_date": {
            "description": "A date of dissolution",
            "pattern": "^[0-9]{4}(-[0-9]{2}){0,2}$",
            "type": ["string", "null"],
        },
        "founding_date": {
            "description": "A date of founding",
            "pattern": "^[0-9]{4}(-[0-9]{2}){0,2}$",
            "type": ["string", "null"],
        },

        "updated_at": {
            "description": "The time at which the resource was last modified",
            "type": ["string", "datetime", "null"],
        },
        "created_at": {
            "description": "The time at which the resource was created",
            "type": ["string", "datetime", "null"],
        },

        "identifiers": identifiers,
        "other_names": other_names,
        "contact_details": contact_details,
        "links": links,

        "name": {
            "description": "A primary name, e.g. a legally recognized name",
            "type": "string"
        },
        "parent_id": {
            "description": "The ID of the organization that contains this organization",
            "type": ["string", "null"],
        },
        "image": {
            "description": "A URL of an image",
            "type": ["string", "null"],
        },
        "posts": {
            "description": "Posts within the organization",
            "items": {
                "properties": {
                    "contact_details": contact_details,
                    "links": links,
                    "id": {
                        "description": "The post's unique identifier",
                        "type": ["string", "null"],
                    },
                    "label": {
                        "description": "A label describing the post",
                        "type": "string"
                    },
                    "role": {
                        "description": "The function that the holder of the post fulfills",
                        "type": ["string", "null"],
                    },
                    "start_date": {
                        "description": "The date on which the post was created",
                        "pattern": "^[0-9]{4}(-[0-9]{2}){0,2}$",
                        "type": ["string", "null"],
                    },
                    "end_date": {
                        "description": "The date on which the post was eliminated",
                        "pattern": "^[0-9]{4}(-[0-9]{2}){0,2}$",
                        "type": ["string", "null"],
                    },
                },
                "title": "Post",
                "type": "object"
            },
            "type": "array"
        },
        "sources": sources,
    },
    "title": "Organization",
    "type": "object",
    "_order": (
        ('Basics', ('name', 'classification', 'parent_id', 'contact_details', 'links')),
        ('Posts', ('posts',)),
        ('Extended Details', ('founding_date', 'dissolution_date',)),
        ('Alternate Names/Identifiers', ('identifiers', 'other_names')),
        ('Common Fields', ('updated_at', 'created_at', 'sources')),
    ),
}



"""
    Schema for bill objects.
"""

from .common import sources


BILL_TYPES = ['bill', 'resolution', 'concurrent resolution',
              'joint resolution', 'memorial']
ACTION_TYPES = ['introduced', 'reading:1', 'reading:2', 'reading:3']
VERSION_TYPES = []
DOCUMENT_TYPES = []


schema = {
    "description": "bill data",
    "type": "object",
    "_order": (
        ('Basics', ('_type', 'organization', 'organization_id', 'session', 'name', 'chamber',
                    'title', 'type', 'subject', 'summaries')),
        ('Common Fields', ['updated_at', 'created_at', 'sources']),
        ('Other/Related Bills', ('other_titles', 'other_names', 'related_bills')),
        ('Sponsors and Actions', ('sponsors', 'actions')),
        ('Documents and Versions', ('documents', 'versions')),
    ),
    "properties": {

        "_type": {"enum": ["bill"], "type": "string",
                  "description": "All bills have a _type field set to bill."
                 },

        "organization": {"type": ["string", "null"],
                         "description": "name of the legislative body that this bill belongs to",
                        },

        "organization_id": {"type": ["string", "null"],
                            "description": "ID of legislative body that this bill belongs to",
                           },

        "session": {"type": "string",
                    "description": "associated with one of the jurisdiction's sessions"},

        "name": {"type": "string",
                 "description": ("jurisdiction-assigned permanent name.  Must be unique within a "
                                 "given session (e.g. HB 3).  Note: not to be confused with "
                                 "``title``.")
                },

        'updated_at': {"type": ["string", "datetime"], "required": False,
                    "description": "the time that the object was last updated",
                   },

        'created_at': {"type": ["string", "datetime"], "required": False,
                    "description": "the time that this object was first created" },

        'chamber': {
            "enum": ["upper", "lower", "joint"], "type": ["string", "null"],
            "description": "chamber vote took place in (if legislature is bicameral, otherwise null)",
        },

        "title": {"type": "string", "description": "primary display title for the bill"},

        "type": {"items": {"type": "string"}, "type": "array",
                 "description": "array of types (e.g. bill, resolution)"},

        "subject": {
            "items": {"type": "string"},
            "type": "array",
            "description": "List of related subjects.",
        },

        "summaries": {
            "items": {
                "properties": {
                    "text": {"type": "string", "description": "Summary of bill."},
                    "note": {"type": ["string", "null"],
                                      "description": "note describing source of summary"},
                },
                "type": "object"
            },
            "type": "array",
            "description": ("List of summaries of bill, each item in list has a note and text "
                            "attribute."),
        },

        "other_titles": {
            "items": {
                "properties": {
                    "title": {"type": "string", "description": "Alternate title."},
                    "note": {"type": ["string", "null"], "description": "Note describing source."},
                },
                "type": "object"
            },
            "type": "array",
            "description": ("list of other titles this bill is known by."
                            "A common use is when a state provides a common title and a long "
                            "or technical title as well.  It is also acceptable to include "
                            "popular but unofficial titles of the bill as well, such as"
                            "'Obamacare' for the 'Patient Protection and Affordable Care Act.'"
                            "`note` can be used to describe the relationship this has to the "
                            "bill, for example Obamacare might be noted as a colloquial name. "
                            "Each item in the list has a title and a note.")
        },

        "other_names": {
            "items": {
                "properties": {
                    "name": {"type": "string", "description": "name (e.g. HB 22)"},
                    "note": {"type": ["string", "null"],
                             "description": "note describing why this name is attached"},
                },
                "type": "object"
            },
            "type": "array",
            "description": ("list of other names this bill is known by in the current session, "
                            "for example if HB 33 and SB 17 refer to the same bill this prevents "
                            "having to have identical entries for each.")

        },

        "related_bills": {
            "items": {
                "properties": {
                    "session": {"type": "string",
                                "description": "Session of related bill."
                               },
                    "name": {"type": "string",
                             "description": "Name of related bill."
                            },
                    "relation_type": {
                        "enum": [
                            "companion",
                            "other-session",
                        ],
                        "type": "string",
                        "description": (
                            "Currently should be either 'companion',"
                            "or 'other-session' for reintroductions"
                        ),
                    },
                },
                "type": "object"
            },
            "type": "array",
            "description": ("Links to related bills.  Currently only used for companion bills, "
                            "but extensible for other uses."),

        },

        "sponsors": {
            "items": {
                "properties": {
                    "sponsorship_type": {"type": "string",
                                         "description": "Type of sponsorship, via upstream source."
                                        },
                    "primary": {
                        "type": "boolean",
                        "description": "Indicates if sponsor is considered primary by source",
                    },
                    "name": {"type": "string",
                             "description": "Name of sponsor, as given by source.",
                            },
                    # TODO: convert to hint object?
                    "chamber": {"enum": ["upper", "lower"], "type": ["string", "null"],
                                "description": "Chamber of sponsor.",
                               },
                    "id": {"type": ["string", "null"],
                           "description": ("ID of entity if the sponsor has been resolved to "
                                           "another entity in the database.")
                          },
                    "_type": {"type": ["string", "null"], "enum": ["organization", "person"],
                           "description": ("Type of entity if the sponsor has been resolved to "
                                           "another entity in the database.")
                             }
                },
                "type": "object"
            },
            "type": "array",
            "description": "List of entities responsible for sponsoring/authoring the bill."
        },

        "actions": {
            "items": {
                "properties": {
                    "description": {
                        "type": "string",
                        "description": "description of the action taken as given by source"
                    },
                    "actor": {
                        "type": ["string", "null"],
                        "description": "name for the actor (e.g. 'upper', 'lower', etc.)"
                    },
                    "date": {
                        "pattern": "^[0-9]{4}(-[0-9]{2}){0,2}$", "type": "string",
                        "description": "date of action"
                    },
                    "type": {
                        "items": {
                            "type": "string"
                        },
                        "type": "array",
                        "description": "array of normalized action types",
                    },
                    "related_entities": {
                        "items": {
                            "properties": {
                                "name": {"type": "string",
                                         "description": "Name of entity given by source data"},
                                "_type": {"enum": ["organization", "person"],
                                          "type": ["string", "null"],
                                          "description": ("Type of entity if the sponsor has been "
                                                          "resolved to another entity in the "
                                                          "database."),

                                         },
                                "id": {"type": ["string", "null"],
                                       "description": ("ID of entity if the sponsor has been "
                                                       "resolved to another entity in the "
                                                       "database.")
                                      },
                            },
                            "type": "object"
                        },
                        "type": "array",
                        "description": ("list of related entities for the action, such as related "
                                        "committee for a referral or a person for a sponsorship.")
                    },
                },
                "type": "object"
            },
            "type": "array",
            "description": "List of actions taken on the bill."
        },


        # == Documents and Versions ==

        "versions": {
            "items": {
                "properties": {
                    "name": {"type": "string", "description": "Name of version"},
                    "type": {"type": ["string", "null"], "enum": VERSION_TYPES,
                             "description": "Type of version"},
                    "date": {
                        "pattern": "^[0-9]{4}(-[0-9]{2}){0,2}$",
                        "type": ["string", "null"],
                        "description": "Version posting date",
                    },
                    "links": {
                        "items": {
                            "properties": {
                                "mimetype": {"type": ["string", "null"],
                                             "description": "mimetype of document"
                                            },
                                "url": {"type": "string", "description": "URL to document" }
                            },
                            "type": "object"
                        },
                        "type": "array",
                        "description": "List of links for this version (pdf, html, etc.)."
                    },
                },
                "type": "object"
            },
            "type": "array",
            "description": "Versions of a bill's text (First Printing, As Amended, etc.)"
        },
        "documents": {
            "items": {
                "properties": {
                    "name": {"type": "string", "description": "Name of document"},
                    "type": {"type": ["string", "null"], "enum": DOCUMENT_TYPES,
                             "description": "Type of document"},
                    "date": {
                        "pattern": "^[0-9]{4}(-[0-9]{2}){0,2}$",
                        "type": ["string", "null"],
                        "description": "Document posting date",
                    },
                    "links": {
                        "items": {
                            "properties": {
                                "mimetype": {"type": ["string", "null"],
                                             "description": "mimetype of document"
                                            },
                                "url": {"type": "string", "description": "URL to document" }
                            },
                            "type": "object"
                        },
                        "type": "array",
                        "description": "List of links to text for this document (pdf, html, etc.)."
                    },
                },
                "type": "object"
            },
            "type": "array",
            "description": "Any non-version related documents, elements are identical to versions.",
        },
        "sources": sources,
    }
}



from ..event import Event
import datetime as dt


def event_obj():
    e = Event(name="get-together",
              when=dt.datetime.utcnow(),
              location="Joe's Place")

    e.add_source(url='foobar')
    e.validate()
    return e


def test_basic_event():
    """ test that we can create an event """
    e = Event(name="get-together",
              when=dt.datetime.utcnow(),
              location="Joe's Place")

    e.add_source(url='foobar')
    e.validate()

    e.add_link("http://foobar.baz")
    e.add_link("http://foobar.baz", note="foo")
    e.validate()

    assert len(e.links) == 2


def test_basic_agenda():
    e = Event(name="get-together",
              when=dt.datetime.utcnow(),
              location="Joe's Place")

    e.add_source(url='foobar')
    e.validate()

    agenda = e.add_agenda_item("foo bar")
    assert agenda
    e.validate()


def test_add_person():
    e = event_obj()
    agenda = e.add_agenda_item("foo bar")
    assert agenda['related_entities'] == []

    agenda.add_person(person='John Q. Hacker', note='chair')
    e.validate()


def test_add_committee():
    e = event_obj()
    agenda = e.add_agenda_item("foo bar")
    assert agenda['related_entities'] == []

    agenda.add_committee(committee='Hello, World', note='host')
    e.validate()


def test_add_bill():
    e = event_obj()
    agenda = e.add_agenda_item("foo bar")
    assert agenda['related_entities'] == []
    agenda.add_bill(bill='HB 101', note='consideration')
    e.validate()


def test_add_document():
    e = event_obj()
    assert e.documents == []
    e.add_document(name='hello', url='http://example.com',
                   mimetype="text/html")
    assert len(e.documents) == 1
    o = e.documents[0]
    assert o['name'] == 'hello'
    assert o['url'] == 'http://example.com'
    e.validate()


def test_add_media():
    e = event_obj()
    e.validate()
    name = "Hello, World"

    a = e.add_agenda_item(description='foo')

    a.add_media_link(name=name, url="http://pault.ag", type='media',
                     mimetype="text/html")

    a.add_media_link(name=name, url="ftp://pault.ag",
                     type='media', mimetype="text/ftp-or-something")

    e.validate()

    assert len(e.agenda[0]['media']) == 1
    assert len(e.agenda[0]['media'][0]['links']) == 2

    e.add_media_link(name=name, url="http://pault.ag", type='media',
                     mimetype="text/html")

    e.add_media_link(name=name, url="ftp://pault.ag",
                     type='media', mimetype="text/ftp-or-something")

    e.validate()

    assert len(e.media) == 1
    assert len(e.media[0]['links']) == 2



import os
import glob
import json
import uuid
import logging
import datetime
from collections import defaultdict
from pupa.core import db
from pupa.utils.topsort import Network


def _hash(obj):
    """ recursively hash unhashable objects """
    if isinstance(obj, (set, tuple, list)):
        return hash(tuple(_hash(e) for e in o))
    elif isinstance(obj, dict):
        return hash(frozenset((k, _hash(v)) for k, v in obj.items()))
    else:
        return hash(obj)


def make_id(type_):
    return 'ocd-{0}/{1}'.format(type_, uuid.uuid1())


def insert_object(obj):
    """ insert a new object into the appropriate collection

    params:
        obj - object to insert

    return:
        database id of new object
    """
    # XXX: check if object already has an id?

    # add updated_at/created_at timestamp
    obj.updated_at = obj.created_at = datetime.datetime.utcnow()
    obj._id = make_id(obj._type)

    obj.save()
    return obj._id


def update_object(old, new):
    """
        update an existing object with a new one, only saving it and
        setting updated_at if something changed

        params:
            old: old object
            new: new object

        returns:
            database_id     id of object in db
            was_updated     whether or not the object was updated
    """
    updated = False

    if old._type != new._type:
        raise ValueError('old and new must be of same _type')

    # allow objects to prevent certain fields from being updated
    locked_fields = old._meta.get('locked_fields', [])

    for key, value in new.as_dict().items():
        if key in locked_fields or key == '_id':
            continue

        if not hasattr(old, key) or getattr(old, key) != value:
            # If we have a *new* value, let's update.
            setattr(old, key, value)
            updated = True

    if updated:
        old.updated_at = datetime.datetime.utcnow()
        old.save()

    return old._id, updated


class BaseImporter(object):

    def __init__(self, jurisdiction_id):
        self.jurisdiction_id = jurisdiction_id
        self.collection = db[self._model_class._collection]
        self.results = {'insert': 0, 'update': 0, 'noop': 0}
        self.json_to_db_id = {}
        self.logger = logging.getLogger("pupa")
        self.info = self.logger.info
        self.debug = self.logger.debug
        self.warning = self.logger.warning
        self.error = self.logger.error
        self.critical = self.logger.critical

    def import_object(self, obj):
        if isinstance(obj, dict):
            raise ValueError("It appears that we're trying to import a dict.")

        spec = self.get_db_spec(obj)

        db_obj = self.collection.find_one(spec)

        if db_obj:
            db_obj = self._model_class.from_dict(db_obj)
            _id, updated = update_object(db_obj, obj)
            self.results['update' if updated else 'noop'] += 1
        else:
            _id = insert_object(obj)
            self.results['insert'] += 1
        return _id

    def dedupe_json_id(self, jid):
        nid = self.duplicates.get(jid, jid)
        if nid != jid:
            return self.dedupe_json_id(nid)
        return jid

    def import_from_json(self, datadir):
        # load all json, mapped by json_id
        raw_objects = {}
        for fname in glob.glob(os.path.join(datadir, self._type + '_*.json')):
            with open(fname) as f:
                data = json.load(f)
                # prepare object from json
                if data['_type'] != 'person':
                    data['jurisdiction_id'] = self.jurisdiction_id
                data = self.prepare_object_from_json(data)
                # convert dict=>class and store in raw_objects
                obj = self._model_class.from_dict(data)
                json_id = obj._id
                raw_objects[json_id] = obj

        # map duplicate ids to first occurance of same object
        inverse = defaultdict(list)
        for json_id, obj in raw_objects.items():
            inverse[_hash(obj)].append(json_id)

        self.duplicates = {}

        for json_ids in inverse.values():
            for json_id in json_ids[1:]:
                self.duplicates[json_id] = json_ids[0]

        # now do import, ignoring duplicates

        # Firstly, before we start, let's de-dupe the pool.
        import_pool = {k: v for k, v in raw_objects.items() if k not in self.duplicates}

        # Now, we create a pupa.utils.topsort.Network object, so that
        # we can contain the import dependencies.
        network = Network()

        to_import = []  # Used to hold the import order
        seen = set()   # Used to ensure we got all nodes.

        for json_id, obj in import_pool.items():
            parent_id = getattr(obj, 'parent_id', None)
            if parent_id:
                # Right. There's an import dep. We need to add the edge from
                # the parent to the current node, so that we import the parent
                # before the current node.
                network.add_edge(parent_id, json_id)
            else:
                # Otherwise, there is no parent, and we just need to add it to
                # the network to add whenever we feel like it during the import
                # phase.
                network.add_node(json_id)

        for link in network.sort():
            to_import.append((link, import_pool[link]))
            seen.add(link)  # This extra step is to make sure that our plan
            # is actually importing all entries into the database.

        if seen != set(import_pool.keys()):  # If it's gone wrong (shouldn't)
            raise ValueError("""Something went wrong internally with the
                                dependency resolution.""")
            # We'll blow up, since we've not done our job and failed to import
            # all of our files into the Database.

        for json_id, obj in to_import:
            parent_id = getattr(obj, 'parent_id', None)
            if parent_id:
                # If we've got a parent ID, let's resolve it's JSON id
                # (scrape-time) to a Database ID (needs to have had the
                # parent imported first - which we asserted is true via
                # the topological sort)
                obj.parent_id = self.resolve_json_id(parent_id)
            self.json_to_db_id[json_id] = self.import_object(obj)

        return {self._type: self.results}

    def resolve_json_id(self, json_id):
        """
            Given an id found in scraped JSON, return a DB id for the object.

            params:
                json_id:    id from json

            returns:
                database id

            raises:
                ValueError if id couldn't be resolved
        """
        if not json_id:
            return None

        json_id = self.dedupe_json_id(json_id)

        # make sure this sort of looks like a UUID
        if len(json_id) != 36:
            raise ValueError('cannot resolve non-uuid: {0}'.format(json_id))

        try:
            return self.json_to_db_id[json_id]
        except KeyError:
            raise ValueError('cannot resolve id: {0}'.format(json_id))

    def prepare_object_from_json(self, obj):
        # no-op by default
        return obj



from pupa.scrape import Scraper
from pupa.scrape.helpers import Legislator, Organization


class PersonScraper(Scraper):

    def get_people(self):
        # committee
        tech = Organization('Technology', classification='committee')
        tech.add_post('Chairman', 'chairman')
        tech.add_source('https://example.com')
        yield tech

        # subcommittee
        ecom = Organization('Subcommittee on E-Commerce',
                            parent=tech,
                            classification='committee')
        ecom.add_source('https://example.com')
        yield ecom

        p = Legislator('Paul Tagliamonte', '6')
        p.add_membership(tech, role='chairman')
        p.add_source('https://example.com')
        yield p





class BaseCommand(object):

    def __init__(self, subparsers):
        self.subparser = subparsers.add_parser(self.name, description=self.help)
        self.add_args()

    def add_args(self):
        pass

    def add_argument(self, *args, **kwargs):
        self.subparser.add_argument(*args, **kwargs)

    def handle(self, args):
        raise NotImplementedError('commands must implement handle(args)')



import datetime
from pupa.core import db
from pupa.models import Event
from .base import BaseImporter


class EventImporter(BaseImporter):
    _type = 'event'
    _model_class = Event

    def get_db_spec(self, event):
        spec = {
            "description": event.description,
            "when": event.when,
            'jurisdiction_id': event.jurisdiction_id,
        }
        return spec

    def prepare_object_from_json(self, obj):

        def person(obj, what):
            spec = {}
            spec['session'] = obj['session']
            if 'chamber' in what:
                spec['chamber'] = what['chamber']
            spec['name'] = what['name']
            # needs to get the right session (current)
            return spec

        def bill(obj, what):
            spec = {}
            spec['session'] = obj['session']
            if 'chamber' in what:
                spec['chamber'] = what['chamber']
            spec['bill_id'] = what['name']
            # needs to get the right session (current)
            return spec

        def org(obj, what):
            spec = {}
            spec['session'] = obj['session']
            if 'chamber' in what:
                spec['chamber'] = what['chamber']
            spec['name'] = what['name']
            # needs to get the right session (current)
            return spec

        spec_generators = {
            "person": person,
            "bill": bill,
            "organization": org,
        }

        # XXX participants

        for item in obj['agenda']:
            for entity in item['related_entities']:
                handler = spec_generators[entity['type']]
                spec = handler(obj, entity)
                spec['jurisdiction_id'] = obj['jurisdiction_id']
                rel_obj = db.events.find_one(spec)
                if rel_obj:
                    entity['id'] = rel_obj['_id']
                else:
                    self.logger.warning('Unknown related entity: {name} '
                                        '({type})'.format(**entity))

        # update time
        obj['when'] = datetime.datetime.fromtimestamp(obj['when'])
        if obj.get('end'):
            obj['end'] = datetime.datetime.fromtimestamp(obj['end'])
        # TODO: handle timezones better

        return obj



from __future__ import print_function
import itertools
from pupa.core import db
from .base import BaseCommand
from pymongo import ASCENDING, DESCENDING


class Command(BaseCommand):
    name = 'ensure-indexes'
    help = '''make mongodb indexes'''

    def add_args(self):
        self.add_argument('collections', nargs='*',
                          help='collections to index (default: all')
        self.add_argument('--purge', action='store_true', default=False,
                          help='purge old indexes')

    def handle(self, args):
        all_indexes = {
            'jurisdictions': [
                [('name', ASCENDING)]
            ],
            'organizations': [],
            'people': [],
            'memberships': [
                # get all members for an org
                [('organization_id', ASCENDING), ('end_date', ASCENDING)],
                # get all memberships for a person
                [('person_id', ASCENDING), ('end_date', ASCENDING)],
            ],
            'bills': [],
            'events': [],
            'votes': [],
        }
        api_indexes = {
            'organizations': [
                ['jurisdiction_id'],
                ['classification'],
                ['parent_id'],
                ['division_id'],
                ['identifiers.scheme', 'identifiers.identifier'],
                [],
            ],
            'people': [
                ['identifiers.scheme', 'identifiers.identifier'],
                [],
            ],
            'bills': [
                ['jurisdiction_id'],
                ['identifiers.scheme', 'identifiers.identifier'],
                [],
            ],
            'events': [
                ['jurisdiction_id'],
                ['participants.id'],
                ['agenda.related_entities.id'],
                [],
            ],
            'votes': [
                ['jurisdiction_id'],
                ['session'],
                ['bill.id'],
            ],
        }

        for _type, indexes in api_indexes.items():
            for fields in indexes:
                real_index = zip(fields, itertools.repeat(ASCENDING))
                all_indexes[_type].append(real_index +
                                          [('created_at', DESCENDING)])
                all_indexes[_type].append(real_index +
                                          [('updated_at', DESCENDING)])
                if _type == 'events':
                    all_indexes[_type].append(real_index +
                                              [('when', DESCENDING)])
                if _type == 'date':
                    all_indexes[_type].append(real_index +
                                              [('date', DESCENDING)])

        collections = args.collections or all_indexes.keys()

        for collection in collections:
            print('indexing', collection, '...')
            current = set(db[collection].index_information().keys())
            current.discard('_id_')
            #if collection == 'bills':
                # basic lookup / unique constraint on abbr/session/bill_id
                #current.discard('%s_1_session_1_chamber_1_bill_id_1' %
                #                settings.LEVEL_FIELD)
                #db.bills.ensure_index([
                #    (settings.LEVEL_FIELD, pymongo.ASCENDING),
                #    ('session', pymongo.ASCENDING),
                #    ('chamber', pymongo.ASCENDING),
                #    ('bill_id', pymongo.ASCENDING)
                #], unique=True)
                #print('creating level-session-chamber-bill_id index')
            print('currently has', len(current), 'indexes (not counting _id)')
            print('ensuring', len(all_indexes[collection]), 'indexes')
            ensured = set()
            for index in all_indexes[collection]:
                if isinstance(index, list):
                    ensured.add(db[collection].ensure_index(index))
                elif isinstance(index, dict):
                    name, index_spec = index.items()[0]
                    ensured.add(
                        db[collection].ensure_index(index_spec, name=name))
                else:
                    raise ValueError(index)
            new = ensured - current
            old = current - ensured
            if len(new):
                print(len(new), 'new indexes:', ', '.join(new))
            if len(old):
                print(len(old), 'indexes deprecated:', ', '.join(old))
                if args.purge:
                    print('removing deprecated indexes...')
                    for index in old:
                        db[collection].drop_index(index)



MONGO_HOST = 'localhost'
MONGO_PORT = 27017
MONGO_DATABASE = 'opencivicdata'

SCRAPELIB_RPM = 60
SCRAPELIB_TIMEOUT = 60
SCRAPELIB_RETRY_ATTEMPTS = 3
SCRAPELIB_RETRY_WAIT_SECONDS = 20

ENABLE_ELASTICSEARCH = False
ELASTICSEARCH_HOST = 'localhost'
ELASTICSEARCH_TIMEOUT = 2

BILL_FILTERS = {}
LEGISLATOR_FILTERS = {}
EVENT_FILTERS = {}

LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': "%(asctime)s %(levelname)s %(name)s: %(message)s",
            'datefmt': '%H:%M:%S'
        }
    },
    'handlers': {
        'default': {'level': 'DEBUG',
                    'class': 'pupa.ext.ansistrm.ColorizingStreamHandler',
                    'formatter': 'standard'},
    },
    'loggers': {
        '': {
            'handlers': ['default'], 'level': 'DEBUG', 'propagate': True
        },
        'scrapelib': {
            'handlers': ['default'], 'level': 'INFO', 'propagate': False
        },
        'requests': {
            'handlers': ['default'], 'level': 'WARN', 'propagate': False
        },
        'boto': {
            'handlers': ['default'], 'level': 'WARN', 'propagate': False
        },
    },
}



from __future__ import print_function
import os
import sys
import glob
import importlib
import warnings
import traceback
from collections import defaultdict
from functools import reduce

from .base import BaseCommand
from pupa import utils

from pupa.importers.jurisdiction import import_jurisdiction
from pupa.importers.organizations import OrganizationImporter
from pupa.importers.people import PersonImporter
from pupa.importers.memberships import MembershipImporter
from pupa.importers.events import EventImporter
from pupa.importers.bills import BillImporter
from pupa.importers.votes import VoteImporter
from pupa.scrape import Jurisdiction


ALL_ACTIONS = ('scrape', 'import', 'report')


class UpdateError(Exception):
    pass


class Command(BaseCommand):
    name = 'update'
    help = 'update pupa data'

    def add_args(self):
        # what to scrape
        self.add_argument('module', type=str, help='path to scraper module')
        for arg in ALL_ACTIONS:
            self.add_argument('--' + arg, dest='actions',
                              action='append_const', const=arg,
                              help='only run {0} post-scrape step'.format(arg))
        for arg in ('people', 'bills', 'events', 'votes', 'speeches'):
            self.add_argument('--' + arg, dest='scrapers',
                              action='append_const', const=arg,
                              help='run {0} scraper'.format(arg))

        self.add_argument('-s', '--session', action='append', dest='sessions',
                          default=[], help='session(s) to scrape')
        self.add_argument('-t', '--term', dest='term', help='term to scrape')

        # debugging
        self.add_argument('--debug', nargs='?', const='pdb', default=None,
                          help='drop into pdb (or set =ipdb =pudb)')

        # scraper arguments
        self.add_argument('--datadir', help='data directory',
                          default=os.path.join(os.getcwd(), 'scraped_data'))
        self.add_argument('--cachedir', help='cache directory',
                          default=os.path.join(os.getcwd(), 'scrape_cache'))
        self.add_argument('--nonstrict', action='store_false', dest='strict',
                          default=True, help='skip validation on save')
        self.add_argument('--fastmode', action='store_true', default=False,
                          help='use cache and turn off throttling')
        self.add_argument('-r', '--rpm', help='scrapelib rpm', type=int,
                          dest='SCRAPELIB_RPM')
        self.add_argument('--timeout', help='scrapelib timeout', type=int,
                          dest='SCRAPELIB_TIMEOUT')
        self.add_argument('--retries', help='scrapelib retries', type=int,
                          dest='SCRAPELIB_RETRIES')
        self.add_argument('--retry_wait', help='scrapelib retry wait',
                          type=int, dest='SCRAPELIB_RETRY_WAIT_SECONDS')

    def enable_debug(self, debug):
        # turn debug on
        if debug:
            _debugger = importlib.import_module(debug)

            # turn on PDB-on-error mode
            # stolen from http://stackoverflow.com/questions/1237379/
            # if this causes problems in interactive mode check that page
            def _tb_info(type, value, tb):
                traceback.print_exception(type, value, tb)
                _debugger.pm()
            sys.excepthook = _tb_info

    def get_jurisdiction(self, module_name):
        # get the jurisdiction object
        module = importlib.import_module(module_name)
        for obj in module.__dict__.values():
            if (getattr(obj, 'jurisdiction_id', None) and
                    issubclass(obj, Jurisdiction)):
                # In the case of top-level scraper classes, the scraper objects
                # will have a jurisdiction_id, but not actually be
                # jurisdictions. As a result, we should ensure all objects we
                # pick up are actually Jurisdiction subclasses.

                # instantiate the class
                retobj = obj()
                if hasattr(retobj, 'get_metadata'):
                    warnings.warn('Jurisdiction class uses deprecated old-style get_metadata')
                    metadata = retobj.get_metadata()
                    retobj.name = metadata['name']
                    retobj.url = metadata['url']
                    retobj.terms = metadata['terms']
                    retobj.provides = metadata['provides']
                    retobj.parties = metadata['parties']
                    retobj.session_details = metadata['session_details']
                    retobj.feature_flags = metadata['feature_flags']
                return retobj

        raise UpdateError('unable to import Jurisdiction subclass from ' +
                          module_name)

    def get_timespan(self, juris, term, sessions):
        if term and sessions:
            raise UpdateError('cannot specify both --term and --session')
        elif sessions:
            terms = set()
            for sess in sessions:
                terms.add(juris.term_for_session(sess))
            if len(terms) != 1:
                raise UpdateError('cannot scrape sessions across terms')
            term = terms.pop()
        elif term:
            sessions = juris.get_term_details(term)['sessions']
        else:
            term = juris.terms[-1]['name']
            sessions = juris.terms[-1]['sessions']

        return term, sessions

    def do_scrape(self, juris, args):
        # make output and cache dirs
        utils.makedirs(args.cachedir)
        utils.makedirs(args.datadir)
        # clear json from data dir
        for f in glob.glob(args.datadir + '/*.json'):
            os.remove(f)

        report = {}

        # run scrapers
        for session in args.sessions:
            # get mapping of ScraperClass -> scrapers
            session_scrapers = defaultdict(list)
            for scraper_type in args.scrapers:
                ScraperCls = juris.get_scraper(args.term, session,
                                               scraper_type)
                if not ScraperCls:
                    raise Exception('no scraper for term={0} session={1} '
                                    'type={2}'.format(args.term, session,
                                                      scraper_type))
                session_scrapers[ScraperCls].append(scraper_type)

            report[session] = {}

            # run each scraper once
            for ScraperCls, scraper_types in session_scrapers.items():
                scraper = ScraperCls(juris, session, args.datadir,
                                     args.cachedir, args.strict,
                                     args.fastmode)
                if 'people' in scraper_types:
                    report[session].update(scraper.scrape_people())
                elif 'bills' in scraper_types:
                    report[session].update(scraper.scrape_bills())
                elif 'events' in scraper_types:
                    report[session].update(scraper.scrape_events())
                elif 'votes' in scraper_types:
                    report[session].update(scraper.scrape_votes())
                elif 'speeches' in scraper_types:
                    report[session].update(scraper.scrape_speeches())

        return report

    def do_import(self, juris, args):
        org_importer = OrganizationImporter(juris.jurisdiction_id)
        person_importer = PersonImporter(juris.jurisdiction_id)

        membership_importer = MembershipImporter(juris.jurisdiction_id,
                                                 person_importer,
                                                 org_importer)

        bill_importer = BillImporter(juris.jurisdiction_id,
                                     org_importer)

        vote_importer = VoteImporter(juris.jurisdiction_id,
                                     person_importer,
                                     org_importer,
                                     bill_importer)

        event_importer = EventImporter(juris.jurisdiction_id)

        report = {}

        # TODO: jurisdiction into report
        import_jurisdiction(org_importer, juris)
        report.update(org_importer.import_from_json(args.datadir))
        report.update(person_importer.import_from_json(args.datadir))
        report.update(membership_importer.import_from_json(args.datadir))
        report.update(bill_importer.import_from_json(args.datadir))
        report.update(event_importer.import_from_json(args.datadir))
        report.update(vote_importer.import_from_json(args.datadir))

        return report

    def check_session_list(self, juris):
        sessions = juris.scrape_session_list()

        all_sessions_in_terms = list(
            reduce(lambda x, y: x + y, [x['sessions'] for x in juris.terms])
        )
        # copy the list to avoid modifying it
        session_details = list(juris._ignored_scraped_sessions)

        for k, v in juris.session_details.items():
            try:
                all_sessions_in_terms.remove(k)
            except ValueError:
                raise UpdateError('session {0} exists in session_details but '
                                  'not in a term'.format(k))

            session_details.append(v.get('_scraped_name'))

        if not sessions:
            raise UpdateError('no sessions from scrape_session_list()')

        if all_sessions_in_terms:
            raise UpdateError('no session_details for session(s): %r' %
                              all_sessions_in_terms)

        unaccounted_sessions = []
        for s in sessions:
            if s not in session_details:
                unaccounted_sessions.append(s)
        if unaccounted_sessions:
            raise UpdateError('session(s) unaccounted for: %r' %
                              unaccounted_sessions)

    def handle(self, args):
        self.enable_debug(args.debug)

        juris = self.get_jurisdiction(args.module)

        # modify args in-place so we can pass them around
        if not args.actions:
            args.actions = ALL_ACTIONS
        args.cachedir = os.path.join(args.cachedir, juris.jurisdiction_id)
        args.datadir = os.path.join(args.datadir, juris.jurisdiction_id)

        # terms, sessions, and object types
        args.term, args.sessions = self.get_timespan(juris, args.term,
                                                     args.sessions)
        args.scrapers = args.scrapers or juris.provides

        # print the plan
        print('module:', args.module)
        print('actions:', ', '.join(args.actions))
        print('term:', args.term)
        print('sessions:', ', '.join(args.sessions))
        print('scrapers:', ', '.join(args.scrapers))
        plan = {'module': args.module, 'actions': args.actions,
                'term': args.term, 'sessions': args.sessions,
                'scrapers': args.scrapers}

        report = {'plan': plan}
        if 'scrape' in args.actions:
            self.check_session_list(juris)
            report['scrape'] = self.do_scrape(juris, args)
        if 'import' in args.actions:
            report['import'] = self.do_import(juris, args)

        report['success'] = True

        # XXX: save report instead of printing
        print(report)
        return report



from ..person import Person
from nose.tools import raises
from validictory import ValidationError


def test_basic_invalid_person():
    """ Test that we can create an invalid person, and validation will fail """
    bob = Person("Bob B. Johnson")
    bob.add_source(url='foo')
    bob.validate()

    bob.name = None

    @raises(ValidationError)
    def _():
        bob.validate()
    _()


def test_str():
    """ test __str__ method """
    assert str(Person("Bob B. Johnson")) == "Bob B. Johnson"


def test_magic_methods():
    """ Test the magic methods work """
    bob = Person("John Q. Public, Esq.",
                 gender="male", image="http://example.com/john.jpg",
                 summary="Some person")
    bob.add_source(url='foo')
    bob.validate()

    bob.add_link("http://twitter.com/ev", "Twitter Account")

    assert bob.links == [
        {"note": "Twitter Account",
         "url": "http://twitter.com/ev"}
    ]

    bob.add_name("Thiston", note="What my friends call me")

    assert bob.other_names == [
        {"name": "Thiston",
         "note": "What my friends call me"}
    ]

    bob.add_name("Johnseph Q. Publico",
                 note="Birth name",
                 start_date="1920-01",
                 end_date="1949-12-31")

    assert bob.other_names == [
        {"name": "Thiston",
         "note": "What my friends call me"},
        {"name": "Johnseph Q. Publico",
         "note": "Birth name",
         "start_date": "1920-01",
         "end_date": "1949-12-31"}
    ]


def test_add_contact_information():
    """ Test that we can add contact information """
    bob = Person("John Q. Public, Esq.",
                 gender="male", image="http://example.com/john.jpg",
                 summary="Some person")
    bob.add_source(url='foo')
    bob.validate()

    bob.add_contact_detail(type='voice',
                           value='876-5309',
                           note='Jenny Cell')

    bob.validate()



from pupa.scrape import Jurisdiction

from .people import PersonScraper


class Example(Jurisdiction):
    jurisdiction_id = 'ocd-jurisdiction/country:us/state:ex/place:example'
    name = 'Example Legislature'
    url = 'http://example.com'
    terms = [{
        'name': '2013-2014',
        'sessions': ['2013'],
        'start_year': 2013,
        'end_year': 2014
    }]
    provides = ['people']
    parties = [
        {'name': 'Independent' },
        {'name': 'Green' },
        {'name': 'Bull-Moose'}
    ]
    session_details = {
        '2013': {'_scraped_name': '2013'}
    }

    def get_scraper(self, term, session, scraper_type):
        if scraper_type == 'people':
            return PersonScraper

    def scrape_session_list(self):
        return ['2013']




from ..bill import Bill
from nose.tools import raises
from validictory import ValidationError


def toy_bill():
    b = Bill(name="HB 2017",
             session="2012A",
             title="A bill for an act to raise the cookie budget by 200%",
             organization="Foo Senate",
             type="bill")
    b.add_source("http://uri.example.com/", note="foo")
    b.validate()
    return b


@raises(ValidationError)
def test_basic_invalid_bill():
    """ Test that we can create an invalid bill, and validation will fail """
    b = toy_bill()
    b.name = None
    b.validate()


def test_verify_actions():
    """ Make sure actions work """
    b = toy_bill()
    b.add_action("Some dude liked it.", "some dude", "2013-04-29")
    b.validate()
    # XXX: Check output


def test_verify_related_bill():
    """ Make sure related bills work """
    b = toy_bill()
    b.add_related_bill(name="HB 2020",
                       session="2011A",
                       chamber="upper",
                       relation="companion")  # continuation?
    b.validate()


def test_verify_documents():
    """ Make sure we can add documents """
    b = toy_bill()
    b.add_document_link(name="Fiscal Impact",
                        date="2013-04",
                        url='http://foo.bar.baz')

    b.add_document_link(name="Fiscal Impact",
                        date="2013-04",
                        url="http://hi.example.com/foo#bar",
                        mimetype="text/html")

    b.add_document_link(name="Fiscal Impact",
                        date="2013-04",
                        url='http://foobar.baz')

    b.validate()

    b.add_document_link(name="Fiscal Impact",
                        date="2013-04",
                        url=None,
                        mimetype='foo')

    @raises(ValidationError)
    def bval():
        b.validate()
    bval()


def test_verify_sponsors():
    """ Make sure sponsors work """
    b = toy_bill()
    b.add_sponsor(name="Joe Bleu",
                  sponsorship_type="Author",
                  entity_type="person",
                  primary=True,
                  chamber="upper")
    b.validate()


def test_subjects():
    """ Test we can add subjects """
    b = toy_bill()
    b.add_subject("Foo")
    b.add_subject("Bar")
    b.validate()


def test_versions():
    """ Test that versions work right """
    b = toy_bill()

    b.add_version_link(
        url="http://pault.ag/",
        name="Final Version",
        date="2013-04",
    )

    b.add_version_link(
        url="http://pault.ag/foo",
        name="Final Version",
        date="2013-04",
    )
    b.validate()

    assert len(b.versions) == 1
    assert len(b.versions[0]['links']) == 2

    try:
        b.add_version_link(
            url="http://pault.ag/foo",
            name="Final Version",
            date="2013-04",
        )
        assert True is False, "We didn't break."
    except ValueError:
        pass

    b.add_version_link(
        url="http://pault.ag/foo",
        name="Final Version",
        date="2013-04",
        on_duplicate='ignore'
    )

    assert len(b.versions) == 1
    assert len(b.versions[0]['links']) == 2

    try:
        b.add_version_link(
            url="http://pault.ag/foo",
            name="Finals Versions",
            date="2013-04",
        )
        assert True is False, "We didn't break."
    except ValueError:
        pass

    assert len(b.versions) == 1
    assert len(b.versions[0]['links']) == 2

    b.add_version_link(
        url="http://pault.ag/foovbar",
        name="Finals Versions",
        date="2013-04",
    )

    assert len(b.versions) == 2
    assert len(b.versions[1]['links']) == 1

    b.validate()



import mock
from nose.tools import assert_equal, assert_in, assert_raises

from pupa.models.person import Person
from pupa.models.organization import Organization
from pupa.scrape.base import Scraper


def test_save_object_basics():
    s = Scraper('jurisdiction', '2013', '/tmp/')
    p = Person('Michael Jordan')
    p.add_source('http://example.com')

    with mock.patch('json.dump') as json_dump:
        s.save_object(p)

    # saved in right place
    filename = 'person_' + p._id + '.json'
    assert_in(filename, s.output_names['person'])
    json_dump.assert_called_once_with(p.as_dict(), mock.ANY, cls=mock.ANY)


def test_save_invalid_object():
    s = Scraper('jurisdiction', '2013', '/tmp/')
    p = Person('Michael Jordan')
    # no source, won't validate

    # this is hideous...
    with assert_raises(ValueError):
        s.save_object(p)


def test_save_related():
    s = Scraper('jurisdiction', '2013', '/tmp/')
    p = Person('Michael Jordan')
    p.add_source('http://example.com')
    o = Organization('Chicago Bulls')
    o.add_source('http://example.com')
    p._related.append(o)

    with mock.patch('json.dump') as json_dump:
        s.save_object(p)

    assert_equal(json_dump.mock_calls, [
        mock.call(p.as_dict(), mock.ANY, cls=mock.ANY),
        mock.call(o.as_dict(), mock.ANY, cls=mock.ANY)
    ])



from nose.tools import raises
from ..membership import Membership


@raises(ValueError)
def test_basic_invalid_membership():
    """ Make sure that we can create an invalid membership and break """
    membership = Membership("person_id", "orga_id")
    membership.validate()

    membership.person_id = 33
    membership.validate()


def test_contact_details_on_membership():
    """ Ensure contact details work for memberships """
    membership = Membership("person_id", "orga_id")
    membership.validate()

    membership.add_contact_detail(type='foo',
                                  value='bar',
                                  note='baz')
    membership.validate()



from pupa.core import db
from pupa.utils import fix_bill_id
from pupa.models import Bill
from .base import BaseImporter


class BillImporter(BaseImporter):
    _type = 'bill'
    _model_class = Bill

    def __init__(self, jurisdiction_id, org_importer):
        super(BillImporter, self).__init__(jurisdiction_id)
        self.org_importer = org_importer

    def get_db_spec(self, bill):
        spec = {'jurisdiction_id': bill.jurisdiction_id,
                'session': bill.session,
                'name': bill.name,}

        if hasattr(bill, 'chamber') and bill.chamber is not None:
            spec['chamber'] = bill.chamber

        return spec

    def prepare_object_from_json(self, obj):
        obj['name'] = fix_bill_id(obj['name'])
        # obj['organization']
        org = self.org_importer._resolve_org_by_chamber(self.jurisdiction_id,
                                                        obj['organization'])

        #self.debug("Scraped chamber for %s was `%s'" % (obj['name'],
        #                                                obj['organization']))
        obj['organization'] = org['_id']

        if 'alternate_bill_ids' in obj:
            obj['alternate_bill_ids'] = [fix_bill_id(bid) for bid in
                                         obj['alternate_bill_ids']]

        # XXX: subject categorizer
        # XXX: action categorizer

        for rel in obj['related_bills']:
            rel['bill_id'] = fix_bill_id(rel['bill_id'])
            spec = rel.copy()
            spec['jurisdiction_id'] = obj['jurisdiction_id']
            rel_obj = db.bills.find_one(spec)
            if rel_obj:
                rel['internal_id'] = rel_obj['_id']
            else:
                self.logger.warning('Unknown related bill: {chamber} '
                                    '{session} {bill_id}'.format(**rel))

        return obj



#!/usr/bin/env python
from setuptools import setup, find_packages
from pupa import __version__

long_description = ''

setup(name='pupa',
      version=__version__,
      packages=find_packages(),
      author='James Turk',
      author_email='jturk@sunlightfoundation.com',
      license='BSD',
      url='http://github.com/opencivicdata/pupa/',
      description='scraping framework for muncipal data',
      long_description=long_description,
      platforms=['any'],
      entry_points='''[console_scripts]
pupa = pupa.cli.__main__:main''',
      install_requires=[
          'six',
          'pymongo>=2.5',
          'scrapelib>=0.8',
          'validictory>=0.9',
      ]
)



schema = {
    "description": "Information about a jurisdiction, including session, term, etc. details.",
    "type": "object",
    "_order": (
        ('Basic Details', ('name', 'url', 'chambers', 'terms', 'session_details')),
        ('Additional Metadata', ('feature_flags', 'building_maps')),
    ),
    "properties": {
        "name": {"type": "string",
                 "description": "Name of jurisdiction (e.g. North Carolina General Assembly)"},
        "url": {"type": "string",
                "description": "URL pointing to jurisdiction's website."
               },
        "chambers": {
            "additionalProperties": {
                "properties": {
                    "name": {
                        "type": "string",
                        "description": "Human-readable name of chamber (e.g. Senate)"
                    },
                    "title": {
                        "type": "string",
                        "description": "Title of an individual in this chamber (e.g. Senator)"
                    }
                },
                "type": "object"
            },
            "type": "object",
            "description": ("Dictionary where keys are slugs for chambers (e.g. upper, lower) "
                            " and values describe the chamber in human-readable terms. "
                            " (only needs to be specified if there are multiple chambers)")

        },
        "terms": {
            "type": "array", "minItems": 1, "items": { "type":"object", "properties": {
                "name": {
                    "type": "string",
                    "description": ("Name of term, typically a year span (e.g. 2011-2012)")
                },
                "start_year": {
                    "type": "integer", "minimum": 1000, "maximum": 2020,
                    "description": "Year that term started.",
                },
                "end_year": {
                    "type": "integer", "minimum": 1000, "maximum": 2030,
                    "description": "Year that term ended."
                },
                "sessions": {
                    "type": "array", "minItems": 1, "items": {"type": "string"},
                    "description": ("List of sessions within this term. "
                                    "Each session must also appear in session_details"),
                }
            }},
            "description": "List of all terms, in chronological order.",
        },
        "session_details": {
            "type": "object",
            "additionalProperties": { "type": "object",
            "properties": {
              "type": {"type": "string", "required": False,
                       "description": "Type of session: primary or special." },
              "start_date": {"type": ["datetime"], "required": False,
                             "description": "Start date of session."
                            },
              "end_date": {"type": ["datetime"], "required": False,
                           "description": "End date of session."
                          }
            } },
            "description": ("Dictionary describing sessions, each key is a session slug that "
                            "must also appear in one ``sessions`` list in ``terms``.  Values "
                            "consist of several fields giving more detail about the session.")
        },
        "feature_flags": {
            "type": "array",
            "items": {"type": "string"},
            "description": ("A way to mark certain features as available on a per-jurisdiction "
                            "basis."),
        },
        "building_maps": {
            "type": "array", "items": {"type":"object", "properties": {
                "name": {"type": "string", "description": "Name of map (e.g. Floor 1)"},
                "url": {"type": "string", "description": "URL to map image/PDF"}
            } },
            "description": ("Links to image/PDF maps of the building."),
         }
    }
}
